#------------------------------------------------------------------------------
# snackAmp - imgSlider package 1.0 for snackAmp
# provides slider/scale widget with images for buttons
#
# Music File Player in Tcl/Tk
#
# Copyright (C) 2006 Tom Wilkason
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
#
# Please send any comments/bug reports to
# tom.wilkason@cox.net  (Tom Wilkason)
#------------------------------------------------------------------------------
set todo {
   - rectangles have outline colors
   - look at scale command so we can work directly with seconds
   do we use canvasx and sceenx commands?
   - canvas has its own borderwidth, can use this instead of padding calculations
   - filled background stays dark if the bg is black.
   - images can have anchors, I guess by default NW is used.
   x When initalizing, use the -variable value to set the initial position
   - Create a state machine for the player controls, so they
   look correct (e.g. stop is down if no paused/playing)
}

package require Tk
package require Img
#----------------------------------------------------------------------------
# Timescale namespace
# create the _time scale widget, expand into existing hull
# track the scale factor so we can position the bubble.
#----------------------------------------------------------------------------
namespace eval imgSlider {
}
#----------------------------------------------------------------------------
# Create the slider using a trough and image for the handle
#----------------------------------------------------------------------------
proc imgSlider::create {w args} {
   upvar #0 $w S
   # Save parameters
   set S(trHeight) 5
   set S(padding)  0
   set S(maxVal)   100
   set S(minVal)   0
   set S(ns)       0
   set S(range) [expr {$S(maxVal)-$S(minVal)}]
   foreach {type val} $args {
      switch -glob -- $type {
         "-wid*" {}
         "-imag*" {
            set img $val
            set S(image) $val
         }
         "-to" -
         "-max*" {
            set S(maxVal) $val
            set S(range) [expr {$S(maxVal)-$S(minVal)}]
         }
         "-from" -
         "-min*" {
            set S(minVal) $val
            set S(range) [expr {$S(maxVal)-$S(minVal)}]
         }
         "-troughh*" {
            set S(trHeight) $val
         }
         "-troughc*" {
            set S(trColor) $val
         }
         "-fill*" {
            set S(fillColor) $val
         }
         "-pad*" {
            set S(padding) $val
         }
         "-bg" -
         "-backgr*" {
            set S(bgColor) $val
         }
         "-var*" {
            # set some traces
            set S(variable) $val
            upvar #0 $S(variable) myvar
            set myvar $myvar
            set S(initVar) $myvar
            trace variable $val w [list imgSlider::_varWrite $w]
            trace variable $val r [list imgSlider::_varRead $w]
         }
         "-dep*" {set S(depress) $val}
         "-rel*" {set S(release) $val}
         "-com*" {set S(command) $val}
         default {
            puts "Unknown configuration option $type=$val"
         }
      }
   }
   if {$S(range) <= 0.0} {
      set S(range) 1.0
   }
   # build the canvas
   canvas $w
   foreach {opt key} {-background bgColor} {
      if {[info exists S($key)]} {
         $w configure $opt $S($key)
      }
   }

   bind [winfo parent $w] <Destroy> [list imgSlider::_destroy $w]
   # save parent width
   set S(iwidth) [image width $img]
   set S(maxPixel) [expr {[winfo reqwidth [winfo parent $w]]-$S(iwidth)}]
   #
   # Bubble & filled area
   #
   #          __                 -bubbleTop
   # ========/==\============   0
   #        |    | -------------
   # ========\==/============   trHeight
   #          --                 +r
   set pd $S(padding)
   set radius [expr {[image height $img]/2}]
   set S(yoffset) [expr {$radius - $S(trHeight)/2 + $pd}]
   #
   # Trough (where bubble rolls)
   #
   $w create rectangle 0 $S(yoffset) $S(maxPixel) [expr {$S(yoffset)+$S(trHeight)}] \
         -tag trough -width 0
   foreach {opt key} {-fill trColor} {
      if {[info exists S($key)]} {
         $w itemconfigure trough $opt $S($key)
      }
   }
   #
   $w create image 0 -$S(yoffset) -image $img -anchor nw -tag bubble
   # filled in area
   $w create rectangle 0 $S(yoffset) 0 [expr {$S(yoffset)+$S(trHeight)}] \
         -tag filledTrough  -width 0
   foreach {opt key} {-fill fillColor} {
      if {[info exists S($key)]} {
         $w itemconfigure filledTrough $opt $S($key)
      }
   }

   # Move bubble into view
   $w move bubble 0 [expr {$S(yoffset) + $pd}]

   # Make the canvas fit comfortably around the image and trough
   set bbox [$w bbox all]
   set height [lindex $bbox 3]
   $w config -height [expr {$height + $pd}] -width $S(maxPixel)

   # Bind drag actions, gamma color image
   if {[info exists S(image)]} {
      # allow gamma to work, force single bit transparency
      catch {imageTools::rewrite $S(image)}
      $w bind bubble <Any-Enter> [subst -nocommand {
         $S(image) configure -gamma 1.5
         focus %W
      }]
      $w bind bubble <Any-Leave> [subst -nocommand {
         $S(image) configure -gamma 1
      }]
      $w bind bubble <ButtonPress-1> [subst -nocommand {
         imgSlider::_Scale2Mark %W %x
         $S(image) configure -gamma .5
      }]
      $w bind bubble <ButtonRelease-1> [subst -nocommand {
         $S(image) configure -gamma 1
      }]
   } else {
      $w bind bubble <ButtonPress-1> [subst -nocommand {
         imgSlider::_Scale2Mark %W %x
      }]
   }
   $w bind bubble <B1-Motion> [subst -nocommand {
      imgSlider::_Scale2Drag %W %x
      $S(image) configure -gamma .5
   }]
   # position slider at new mouse position
   $w bind all <ButtonRelease-3> {
      imgSlider::setVar %W [imgSlider::_time %W [imgSlider::_pixToTime %W %x]]
   }
   ##
   # Extra bubble bindings added
   # todo handle passed in bindings, both for bubble and widget itself
   #
   foreach {option} {depress release command} \
         {button} {<ButtonPress-1> <ButtonRelease-1> <B1-Motion>} {
            if {[info exists S($option)]} {
               if {$option eq "command"} {
                  # command passes time value as argument
                  append  S($option) " \[imgSlider::_time $w\]"
               }
         $w bind bubble $button +$S($option)
      }
   }
   # todo, add mousewheel if -increment set to jump forward/backward
   foreach {option} {release} \
         {button} {<ButtonRelease-3>} {
            if {[info exists S($option)]} {
               $w bind all $button +$S($option)
            }
   }
   ##
   # When user expands the hull, expand with it
   $w configure -xscrollcommand [list imgSlider::_reconfigure $w]
   foreach {X Y} [$w coords bubble] {break}
   set S(Y) $Y
   set S(lastPos) 0  ;# last mouse position
   set S(np) 0
   set S(ns) 0
   # Create and return a new command to handle subcommands
   # widget itself isn't renamed (for packing and such)
   set S(widget) ${w}_widget
   rename $w ::${w}_widget
   interp alias {} $w {} imgSlider::_subcommand $w
   ##
   # Set initial value if variable passed in
   #
   if {[info exists S(initVar)]} {
      upvar #0 $S(variable) myvar
      _time $w $S(initVar)
   }
   return $w
}

#------------------------------------------------------------------------------
# Called when widget is destroyed
#------------------------------------------------------------------------------
proc imgSlider::_destroy {w} {
   upvar #0 $w S
   unset -nocomplain S
}
#------------------------------------------------------------------------------
# Handle subcommands
#------------------------------------------------------------------------------
proc imgSlider::_subcommand {w subcommand args} {
   upvar #0 $w S
   switch -glob -- $subcommand {
      "conf*" {
         foreach {type val} $args {
            switch -glob -- $type {
               "-fill*" {
                  set S(fillColor) $val
                  $S(widget) itemconfigure filledTrough -fill $val
               }
               "-imag*" {
                  set img $val
                  set S(image) $val
               }
               "-to" -
               "-max*" {
                  set S(maxVal) $val
                  set S(range) [expr {$S(maxVal)-$S(minVal)}]
               }
               "-from" -
               "-min*" {
                  set S(minVal) $val
                  set S(range) [expr {$S(maxVal)-$S(minVal)}]
               }
               "-troughh*" {
                  set S(trHeight) $val
               }
               "-troughc*" {
                  set S(trColor) $val
                  $S(widget) itemconfigure trough -fill $val
               }
               "-pad*" {
                  set S(padding) $val
               }
               "-bg" -
               "-backgr*" {
                  set S(bgColor) $val
                  $S(widget) configure -background $S(bgColor)
               }
               "-var*" {
                  set S(variable) $val
                  trace variable $val w [list imgSlider::_varWrite $w]
                  trace variable $val r [list imgSlider::_varRead $w]
                  # set some traces
               }
               "-dep*" {$S(widget) bind bubble <ButtonPress-1> [list +$val]}
               "-com*" {$S(widget) bind bubble <B1-Motion> [list +$val]}
               "-rel*" {$S(widget) bind bubble <ButtonRelease-1> [list +$val]}
               default {
                  puts "unknown option $type=$val"
               }
            }
         }
      }
      "set" {
         imgSlider::setVar $w [imgSlider::_time $w [lindex $args 0]]
      }
      "cget" {
         return "xx"
      }
      default {}
   }
   if {$S(range) <= 0.0} {
      set S(range) 1.0
   }
}

#----------------------------------------------------------------------------
# Translate a pixel position to time
#----------------------------------------------------------------------------
proc imgSlider::_pixToTime {w np} {
   upvar #0 $w S
   set ns [expr {double($np)/double($S(maxPixel))*$S(range) + $S(minVal)}]
   return $ns
}
#----------------------------------------------------------------------------
# Initiate a drag
#----------------------------------------------------------------------------
proc imgSlider::_Scale2Mark { w x} {
   upvar #0 $w S
   # Remember the anchor point for the drag
   set S(lastPos) $x
}
#----------------------------------------------------------------------------
# Perform a drag, bound the position
# Have to use a delta method so the mouse doesn't jump to beginning of bubble
#----------------------------------------------------------------------------
proc imgSlider::_Scale2Drag { w x} {
   upvar #0 $w S
   # Compute delta and update anchor point
   foreach {X Y} [$S(widget) coords bubble] {break}
   foreach {X1 Y1 X2 Y2} [$S(widget) coords filledTrough] {break}
   set x1 $S(lastPos) ;#last pos
   set dx [expr {$x - $x1}]
   set re $S(maxPixel)
   # Constrain the movement to be between the
   # left edge and the right edge.
   if {$x < 0 || $x > $re+$S(iwidth)} {
      # ignore if outside area
      return
   } elseif {$X + $dx > $re} {
      # right edge
      set S(lastPos) $re
      set S(np)      $re
      set S(ns) $S(maxVal) ;# seconds position
   } elseif {($X + $dx) < 0} {
      #left edge $X==0  && $dx < 0
      set S(lastPos) $x
      set S(np) 0 ;# pixel position
      set S(ns) $S(minVal) ;# seconds position
   } else {
      # in the middle, get new position and new seconds, then adust bubble
      set np [expr {$X+$dx}]
      set ns [expr {$np/double($re)*$S(range)+$S(minVal)}]
      set S(lastPos) $x
      set S(np) $np ;# pixel position
      set S(ns) $ns ;# seconds position
      $S(widget) coords filledTrough $X1 $Y1 $np $Y2
      #need to move, not locate for mouse
      $S(widget) move bubble $dx 0
   }
}

#----------------------------------------------------------------------------
# Expand the scale to the available space when window resizes
#----------------------------------------------------------------------------
proc imgSlider::_reconfigure {w args} {
   upvar #0 $w S
   set wid [expr {[winfo width $w]-$S(iwidth)}]
   ##
   # Move Slider
   #
   foreach {X1 X2 Y1 Y2} [$S(widget) coords trough] {break}
   $S(widget) coords trough $X1 $X2 $wid $Y2
   ##
   # Move bubble to new scaled position
   #
   foreach {X Y} [$S(widget) coords bubble] {break}
   set ow [expr {double($S(maxPixel))}]
   if {$ow == 0} {
      set ow 30
   }
   set np [expr {round($X * double($wid)/$ow)}]
   set S(maxPixel) $wid
   _pixPosition $w $np
}
#------------------------------------------------------------------------------
# variable trace handlers, on write set the new position
#------------------------------------------------------------------------------
proc imgSlider::_varWrite {w _var index mode args} {
   #Trace
   upvar #0 $w S
   upvar #0 $S(variable) myVar
   # Ignore very small movements
   if {abs($S(ns)-$myVar) > 0.01} {
      set np [expr {round(($myVar-$S(minVal))/double($S(range))*$S(maxPixel))}]
      _pixPosition $w $np
      set S(ns) $myVar
   }
   return $myVar
}
#------------------------------------------------------------------------------
# variable trace handlers, on read return current value
#------------------------------------------------------------------------------
proc imgSlider::_varRead {w _var index mode args} {
   #Trace 1
   upvar #0 $w S
   upvar #0 $S(variable) myVar
   set myVar [imgSlider::_time $w]
   return $myVar
}
#----------------------------------------------------------------------------
# Set the variable to the correct position
#----------------------------------------------------------------------------
proc imgSlider::setVar {w val} {
   upvar #0 $w S
   # Set the linked variable as well (if it exists)
   # kludge, only really needed for initialization
   if {[info exists S(variable)]} {
      upvar #0 $S(variable) myvar
      set myvar $val
   }
}
#----------------------------------------------------------------------------
# Return or set the current values for the trough
#----------------------------------------------------------------------------
proc imgSlider::_time {w {val ""}} {
   #Trace 1
   upvar #0 $w S
   # Return the current values of the double trough
   if {$val eq ""} {
      return $S(ns)
   } else {

      set np [expr {round(($val-$S(minVal))/double($S(range))*$S(maxPixel))}]
      _pixPosition $w $np
      set S(ns) $val
      return $val
   }
}
#----------------------------------------------------------------------------
# Position the bubble to some pixel position
#----------------------------------------------------------------------------
proc imgSlider::_pixPosition {w np} {
   upvar #0 $w S
   set Y $S(Y)
   $S(widget) coords bubble $np $Y
   $S(widget) coords filledTrough 0 $S(yoffset) $np [expr {$S(yoffset)+$S(trHeight)}]
   set S(np) $np
   # also need to move bubble by old x/old width
}

#------------------------------------------------------------------------------
#
#------------------------------------------------------------------------------
proc demo.scale {args} {
   global TA
   set TA(demo) 0
   set ::BUBBA -45
   destroy .demo
   toplevel .demo
   wm geometry .demo 200x50
   set volScale [imgSlider::create .demo.sv \
   -from -80 \
   -to 0 \
   -image saImages.volSlider \
   -background blue \
   -fillcolor  pink \
   -troughheight 4 \
   -padding 2 \
   -command {demo.show} \
   -variable ::BUBBA]
   pack $volScale -fill both -expand true
}

#------------------------------------------------------------------------------
#
#------------------------------------------------------------------------------
proc demo.show {args} {
   global TA
   debug $args
   puts $::BUBBA
}

#------------------------------------------------------------------------------
#
#------------------------------------------------------------------------------
proc show {args} {
   parray ::imgSlider::var *demo*
}
package provide imageScale 1.0


