## Paned Window Procs inspired by code by Stephen Uhler @ Sun.
## Thanks to John Ellson (ellson@lucent.com) for bug reports & code ideas.
##
## Copyright 1996-1997 Jeffrey Hobbs, jeff.hobbs@acm.org
## Modified 2001 Tom Wilkason to support non-balenced pane sizes

##------------------------------------------------------------------
## PROCEDURE
## pane
##
## DESCRIPTION
## paned window management function
##
## METHODS
##
##  pane configure <widget> ?<widget> ...? ?<option> <value>?
##  pane <widget> ?<widget> ...? ?<option> <value>?
## Sets up the management of the named widgets as paned windows.
##
## OPTIONS
## -dynamic Whether to dynamically resize or to resize only
##       when the user lets go of the handle
## -orient     Orientation of window to determing tiling.
##       Can be either horizontal (default) or vertical.
## -parent     A master widget to use for the slaves.
##       Defaults to the parent of the first widget.
## -handlelook Options to pass to the handle during 'frame' creation.
## -handleplace   Options to pass to the handle during 'place'ment.
##       Make sure you know what you're doing.
##
##  pane forget <master> ?<slave> ...?
## If called without a slave name, it forgets all slaves and removes
## all handles, otherwise just removes the named slave(s) and redraws.
##
##  pane info <slave>
## Returns the value of [place info <slave>].
##
##  pane slaves <master>
## Returns the slaves currently managed by <master>.
##
##  pane master <slave>
## Returns the master currently managing <slave>.
##
## BEHAVIORAL NOTES
## pane is a limited interface to paned window management.  Error
##  catching is minimal.  When you add more widgets to an already managed
##  parent, all the fractions are recalculated.  Handles have the name
##  $parent.__h#, and will be created/destroyed automagically.  You must
##  use 'pane forget $parent' to clean up what 'pane' creates, otherwise
##  critical state info about the parent pane will not be deleted.  This
##  could support -before/after without too much effort if the desire
##  was there.  Because this uses 'place', you have to take the same care
##  to size the parent yourself.
##
## VERSION 1.0
##
## EXAMPLES AT END OF FILE
##

## FIX: should be able to set original ratio

proc pane {opt args} {
   global PANE
   switch -glob -- $opt {
      c* { eval pane_config $args }
      f* {
         set p [lindex $args 0]
         if {[info exists PANE($p,w)]} {
            if {[llength $args]==1} {
               foreach w $PANE($p,w) { catch {place forget $w} }
               foreach w [array names PANE $p,*] { unset PANE($w) }
               if {![catch {winfo children $p} kids]} {
                  foreach w $kids {
                     if {[string match *.__h* $w]} { destroy $w }
                  }
               }
            } else {
               foreach w [lrange $args 1 end] {
                  place forget $w
                  set i [lsearch -exact $PANE($p,w) $w]
                  set PANE($p,w) [lreplace $PANE($p,w) $i $i]
               }
               if {[llength $PANE($p,w)]} {
                  eval pane_config $PANE($p,w)
               } else {
                  pane forget $p
               }
            }
         } else {

         }
      }
      i* { return [place info $args] }
      s* {
         if {[info exists PANE($args,w)]} {
            return $PANE($args,w)
         } else {
            return
         }
      }
      m* {
         foreach w [array names PANE *,w] {
            if {[lsearch -exact $PANE($w) $args] != -1} {
               regexp {([^,]*),w} $w . res
               return $res
            }
         }
         return -code error \
               "no master found. perhaps $args is not a pane slave?"
      }
      default { eval pane_config [list $opt] $args }
   }
}

##
## PRIVATE FUNCTIONS
##
## I don't advise playing with these because they are slapped together
## and delicate.  I don't recommend calling them directly either.
##
;proc pane_config {args} {
   global PANE
   array set opt {orn none par {} dyn 0 hpl {} hlk {}}
   set wids {}
   for {set i 0;set num [llength $args];set cargs {}} {$i<$num} {incr i} {
      set arg [lindex $args $i]
      if {[winfo exists $arg]} { lappend wids $arg; continue }
      set val [lindex $args [incr i]]
      switch -glob -- $arg {
         -d*   { set opt(dyn) [regexp -nocase {^(1|yes|true|on)$} $val] }
         -o*   { set opt(orn) $val }
         -p*   { set opt(par) $val }
         -handlep*   { set opt(hpl) $val }
         -handlel*   { set opt(hlk) $val }
         -size {
            ##
            # Set the fractional size for each specified widget
            #
            set index -1
            if {$val == {}} {
               set result [list]
               foreach {widget} $wids {
                  ##
                  # TFW: Want to use array that has current values
                  # not those set by the user
                  #
                  if {[catch {
                     lappend result $PANE($widget,curFrac)
                  }]} {
                     return 0.5
                  }
               }
               return $result
            } else {
               if {[llength $wids] != [llength $val]} {
                  return -code error "one fraction per widget please"
               }
               foreach {widget} $wids {
                  set PANE($widget,frac) [lindex $val [incr index]]
                  set PANE($widget,curFrac) $PANE($widget,frac)
               }
            }
         }
         default  { return -code error "unknown option \"$arg\"" }
      }
   }
   if {[string match {} $wids]} {
      return -code error "no widgets specified to configure"
   }
   if {[string compare {} $opt(par)]} {
      set p $opt(par)
   } else {
      set p [winfo parent [lindex $wids 0]]
   }
   if {[string match none $opt(orn)]} {
      if {![info exists PANE($p,o)]} { set PANE($p,o) h }
   } else {
      set PANE($p,o) $opt(orn)
   }
   if {[string match h* $PANE($p,o)]} {
      set owh height; set wh width; set xy x; set hv h
   } else {
      set owh width; set wh height; set xy y; set hv v
   }
   if {![info exists PANE($p,w)]} { set PANE($p,w) {} }
   foreach w [winfo children $p] {
      if {[string match *.__h* $w]} { destroy $w }
   }
   foreach w $wids {
      set i [lsearch -exact $PANE($p,w) $w]
      if {$i<0} { lappend PANE($p,w) $w }
   }
   ##
   # Determine what fractions are in use
   #
   set fracInUse 0.0
   set ll [llength $PANE($p,w)]
   set noFracSet $ll
   foreach w $PANE($p,w) {
      if {[info exists PANE($w,frac)]} {
         set fracInUse [expr {$fracInUse+ $PANE($w,frac)}]
         set Fraction($w) $PANE($w,frac)
         incr noFracSet -1
      }
   }
   ##
   # Spread the remaining fractions around evenly
   #
   set remFrac [expr {(1.0 - $fracInUse)/$noFracSet}]
   foreach w $PANE($p,w) {
      if {! [info exists PANE($w,frac)]} {
         set Fraction($w) $remFrac
      }
   }
   ##
   # Right now they all get the same size, if this $p has a set
   # fraction then use it.
   #
   set frac [expr {1.0/$ll}]
   set pos 0.0
   array set hndconf $opt(hlk)
   if {![info exists hndconf(-$wh)]} {
      set hndconf(-$wh) 4
   }
   ##
   # frac can be array of same length as $PANE($p,w)
   #
   foreach w $PANE($p,w) {
      # puts "$w $Fraction($w)"
      place forget $w
      place $w -in $p -rel$owh 1 -rel$xy $pos -$wh -$hndconf(-$wh) \
            -rel$wh $Fraction($w) -anchor nw

      raise $w
      set pos [expr {$pos+$Fraction($w)}]
   }
   place $w -$wh 0
   set pos 0.0
   set index -1
   while {[incr ll -1]} {
      set h [eval frame [list $p.__h$ll] -bd 15 -relief raised \
            -cursor sb_${hv}_double_arrow [array get hndconf]]
      ##
      # This just places a bar at n-1/N, we need to put it at
      # the true fractional border
      #
      set W [lindex $PANE($p,w) [incr index]]
      set pos [expr {$pos+$Fraction($W)}]
      eval place [list $h] -rel$owh 1 -rel$xy $pos \
            -$xy -$hndconf(-$wh) -anchor nw $opt(hpl)
      raise $h
      bind $h <ButtonPress-1> [list pane_constrain $p $h \
            [lindex $PANE($p,w) [expr {$ll-1}]] [lindex $PANE($p,w) $ll] \
            $wh $xy $opt(dyn)]
   }
   return 1
}
##
## EXAMPLES
##

;proc pane_constrain {p h w0 w1 wh xy d} {
   global PANE
   regexp -- "\-rel$xy (\[^ \]+)" [place info $w0] junk t0
   regexp -- "\-rel$xy (\[^ \]+).*\-rel$wh (\[^ \]+)" \
         [place info $w1] junk t1 t2
   set offset [expr {($t1+$t2-$t0)/10.0}]
   array set PANE [list XY [winfo root$xy $p] WH [winfo $wh $p].0 \
         W0 $w0 W1 $w1 XY0 $t0 XY1 [expr {$t1+$t2}] \
         C0 [expr {$t0+$offset}] C1 [expr {$t1+$t2-$offset}]]
   bind $h <B1-Motion> [list pane_motion %[string toupper $xy] $p $h $wh $xy $d]
   if {!$d} {
      bind $h <ButtonRelease-1> [list pane_motion %[string toupper $xy] $p $h $wh $xy 1]
   }
}

;proc pane_motion {X p h wh xy d} {
   global PANE
   ##
   # TFW: Want to save the fractions for each of the panes
   # as the default so they can be later resued if
   # p=frame
   set f [expr {($X-$PANE(XY))/$PANE(WH)}]
   # Constrain
   if {$f<$PANE(C0)} { set f $PANE(C0) }
   if {$f>$PANE(C1)} { set f $PANE(C1) }
   # Save fractions

   set PANE([lindex [winfo children $p] 0],curFrac) $f
   set PANE([lindex [winfo children $p] 1],curFrac) [expr {1.0-$f}]
   if {$d} {
      place $PANE(W0) -rel$wh [expr {$f-$PANE(XY0)}]
      place $h -rel$xy $f
      place $PANE(W1) -rel$wh [expr {$PANE(XY1)-$f}] -rel$xy $f
   } else {
      place $h -rel$xy $f
   }
}
package provide saPane 1.0
## EOF


