#------------------------------------------------------------------------------
#!/usr/bin/wish
#
# Copyright (C) 1997,1998 D. Richard Hipp
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Library General Public
# License as published by the Free Software Foundation; either
# version 2 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Library General Public License for more details.
#
# You should have received a copy of the GNU Library General Public
# License along with this library; if not, write to the
# Free Software Foundation, Inc., 59 Temple Place - Suite 330,
# Boston, MA  02111-1307, USA.
#
# Author contact information:
#   drh@acm.org
#   http://www.hwaci.com/drh/
#
# $Revision: 1.22 $
# Modified by Tom Wilkason to add checkboxes for each node if they are linked to a variable.
# Also optimized access to global info and namespaced everything.
# 1/9/2001
# Hack Note: By default a closed folder is the icon, and when selected it turns into an open folder
# This could be changed to pass in openimage, closeimage but for this purpose it saves memory.
# For generalization it should be done this way however.
#
# $Revision: 1.22 $
# Modified by Leslie Brooks
# - Tree:delitem - Added 'catch' around some of the 'unset'; if the tree was
#   not built correctly these array elements might not exist, so trying to
#   delete the element would throw an error.
# - Tree:delitem - converted individual 'unset' lines for each array element
#   to a 'foreach' loop; this will delete everything associated with the node,
#   no matter what elements someone adds in the future.
# - moved initialization of foreground and background colors outside the
#   'switch' statement; they never got executed where they were.
# - Tree:dfltconfig - added code to set the default icon to "" (no icon) if
#   the user did not specify one.
# - Tree:buildlayer - added code to test for the existence of '$icon.close';
#   if it doesn't exist then use '$icon' (which is exactly what the user passed in).
# - Tree:buildlayer - added code to indent properly when the user specifies no icon
# - Tree:toggleNode - added code to toggle a node in State=2 (Partially Selected)
#   to State=1 (Selected).
# - Tree::newitem - added a line to record the parent of each node.  This
#   will make it possible to easily traverse the tree from leaf to root.
# - Tree:buildlayer - changed the code so it draws longer connecting lines
#   if the user specified 'No Icon' for a node.  This allows you to mix nodes
#   that have icons with nodes that don't (possibly leaf nodes) but still
#   keep everything lined up vertically.
# - Tree::dfltconfig - set the default icon to ""; that is, 'No Icon'.
# - Tree::toggleNode - simplified the code, and added code so that any node
#   in State=2 (Partially Selected) will transition to State=1 (Selected)
#   when the user clicks on the check-box.
# - Tree::buildlayer - added code to store the tag associated with each
#   checkbox.  This allows us to change their icons without having the user
#   click on them.  (Required if we are going to change parents and children
#   of the clicked node.)
# - Tree::toggleNode - added code to call Tree::setChildState; this sets the
#   states of all children to the state of the parent.  Created the new proc
#   'Tree:setChildState'.
# - Added isDirty and setDirty methods to access the dirty state which indicates
#   if the checkbox has been changed by the user.
# - Added configure/cget settings to allow the user to setup how the tree
#   should work. Call Tree::configure $w -setting method to use.
#------------------------------------------------------------------------------
#
# Notes:
#   The tree node-names use standard directory syntax - '/foo/fink' refers
# to a node named 'fink' that is a child of 'foo' which is a child of the
# tree root.  Don't use Windows-style name for the nodes.
#   All of the parameters associated with a node (its icon, current state,
# names of its children, etc.) are stored in an array variable named 'Tree',
# and are organized under the canvas name.  So for example 'Tree($w:$v:icon)'
# refers to the icon associated with node $v in tree (canvas, actually) $w.
#
# Common Variable Names:
#   dir - the full path leading to a tree node (for example '/fink/')
#   n - the name of the current tree node (for example, 'foo')
#   v - the fully qualified name of a tree node (for example '/fink/foo')
#   w - the name of the canvas (for example, '.f1.myCanvas')
#
# CONFIGURATION OPTIONS:
#   Call 'Tree::configure $w ?option? ?value' to set configuration options.
#
#   The options are:
#   -rippleSet - Set this option to '1' if you want check-box changes to
#      ripple up and down the tree.  That is, if you select any node, all
#      nodes below it will be selected automatically.  If you de-select
#      a node, all parent nodes will change to 'not selected' or 'partially
#      selected'.  This is the behavior you want if you are going to use
#      the tree for selecting directories and files to back up, or tests
#      to run.  With the default setting of '0', changes to any file or
#      directory have no affect on any other file or directory.
#
#    -treeLine - Set this option to '1' if to show the old windows style
#    +/- boxes and lines between the nodes. Set to '0' if you want a more
#    modern (and mac like) right/down triangle without the lines.
#------------------------------------------------------------------------------

set [file rootname [file tail [info script]]]_History {
   $Header: /cvsroot/snackamp/snackamp/lib/saTree.tcl,v 1.22 2005/12/22 13:46:31 wilkason Exp $
}
##
# Namespace Tree Main, create local variables and default options
#
namespace eval Tree {
   variable fg black
   variable bg white
   variable font
   variable highlight
   variable indent 15
####   option add *highlightThickness 0
   switch $tcl_platform(platform) {
      unix {
         set font \
               -adobe-helvetica-medium-r-normal-*-11-80-100-100-p-56-iso8859-1
      }
      windows {
         set font \
               -adobe-helvetica-medium-r-normal-*-14-100-100-100-p-76-iso8859-1
      }
   }
   # - Box
   image create photo Tree::openbm -data {
      R0lGODlhCQAJAPcAAAAAAAEBAQICAgMDAwQEBAUFBQYGBgcHBwgICAkJCQoK
      CgsLCwwMDA0NDQ4ODg8PDxAQEBERERISEhMTExQUFBUVFRYWFhcXFxgYGBkZ
      GRoaGhsbGxwcHB0dHR4eHh8fHyAgICEhISIiIiMjIyQkJCUlJSYmJicnJygo
      KCkpKSoqKisrKywsLC0tLS4uLi8vLzAwMDExMTIyMjMzMzQ0NDU1NTY2Njc3
      Nzg4ODk5OTo6Ojs7Ozw8PD09PT4+Pj8/P0BAQEFBQUJCQkNDQ0REREVFRUZG
      RkdHR0hISElJSUpKSktLS0xMTE1NTU5OTk9PT1BQUFFRUVJSUlNTU1RUVFVV
      VVZWVldXV1hYWFlZWVpaWltbW1xcXF1dXV5eXl9fX2BgYGFhYWJiYmNjY2Rk
      ZGVlZWZmZmdnZ2hoaGlpaWpqamtra2xsbG1tbW5ubm9vb3BwcHFxcXJycnNz
      c3R0dHV1dXZ2dnd3d3h4eHl5eXp6ent7e3x8fH19fX5+fn9/f4CAgIGBgYKC
      goODg4SEhIWFhYaGhoeHh4iIiImJiYqKiouLi4yMjI2NjY6Ojo+Pj5CQkJGR
      kZKSkpOTk5SUlJWVlZaWlpeXl5iYmJmZmZqampubm5ycnJ2dnZ6enp+fn6Cg
      oKGhoaKioqOjo6SkpKWlpaampqenp6ioqKmpqaqqqqurq6ysrK2tra6urq+v
      r7CwsLGxsbKysrOzs7S0tLW1tba2tre3t7i4uLm5ubq6uru7u7y8vL29vb6+
      vr+/v8DAwMHBwcLCwsPDw8TExMXFxcbGxsfHx8jIyMnJycrKysvLy8zMzM3N
      zc7Ozs/Pz9DQ0NHR0dLS0tPT09TU1NXV1dbW1tfX19jY2NnZ2dra2tvb29zc
      3N3d3d7e3t/f3+Dg4OHh4eLi4uPj4+Tk5OXl5ebm5ufn5+jo6Onp6erq6uvr
      6+zs7O3t7e7u7u/v7/Dw8PHx8fLy8vPz8/T09PX19fb29vf39/j4+Pn5+fr6
      +vv7+/z8/P39/f7+/v///yH5BAAAAAAALAAAAAAJAAkAAAggAAEIHEjwn8GD
      /wQiPKhwYUIADwc+dDjRYcOFBDMCCAgAOw==
   }
   # + Box
   image create photo Tree::closedbm -data {
      R0lGODlhCQAJAPcAAAAAAAEBAQICAgMDAwQEBAUFBQYGBgcHBwgICAkJCQoK
      CgsLCwwMDA0NDQ4ODg8PDxAQEBERERISEhMTExQUFBUVFRYWFhcXFxgYGBkZ
      GRoaGhsbGxwcHB0dHR4eHh8fHyAgICEhISIiIiMjIyQkJCUlJSYmJicnJygo
      KCkpKSoqKisrKywsLC0tLS4uLi8vLzAwMDExMTIyMjMzMzQ0NDU1NTY2Njc3
      Nzg4ODk5OTo6Ojs7Ozw8PD09PT4+Pj8/P0BAQEFBQUJCQkNDQ0REREVFRUZG
      RkdHR0hISElJSUpKSktLS0xMTE1NTU5OTk9PT1BQUFFRUVJSUlNTU1RUVFVV
      VVZWVldXV1hYWFlZWVpaWltbW1xcXF1dXV5eXl9fX2BgYGFhYWJiYmNjY2Rk
      ZGVlZWZmZmdnZ2hoaGlpaWpqamtra2xsbG1tbW5ubm9vb3BwcHFxcXJycnNz
      c3R0dHV1dXZ2dnd3d3h4eHl5eXp6ent7e3x8fH19fX5+fn9/f4CAgIGBgYKC
      goODg4SEhIWFhYaGhoeHh4iIiImJiYqKiouLi4yMjI2NjY6Ojo+Pj5CQkJGR
      kZKSkpOTk5SUlJWVlZaWlpeXl5iYmJmZmZqampubm5ycnJ2dnZ6enp+fn6Cg
      oKGhoaKioqOjo6SkpKWlpaampqenp6ioqKmpqaqqqqurq6ysrK2tra6urq+v
      r7CwsLGxsbKysrOzs7S0tLW1tba2tre3t7i4uLm5ubq6uru7u7y8vL29vb6+
      vr+/v8DAwMHBwcLCwsPDw8TExMXFxcbGxsfHx8jIyMnJycrKysvLy8zMzM3N
      zc7Ozs/Pz9DQ0NHR0dLS0tPT09TU1NXV1dbW1tfX19jY2NnZ2dra2tvb29zc
      3N3d3d7e3t/f3+Dg4OHh4eLi4uPj4+Tk5OXl5ebm5ufn5+jo6Onp6erq6uvr
      6+zs7O3t7e7u7u/v7/Dw8PHx8fLy8vPz8/T09PX19fb29vf39/j4+Pn5+fr6
      +vv7+/z8/P39/f7+/v///yH5BAAAAAAALAAAAAAJAAkAAAghAAEIHEjwn8GD
      /wQeBLCQoUGHCRkSlAixYsWGCBtOHBgQADs=
   }
   # Down arrow
   image create photo Tree::openmac -data {
      R0lGODlhCwAHAPcAAAAAAAEBAQICAgMDAwQEBAUFBQYGBgcHBwgICAkJCQoK
      CgsLCwwMDA0NDQ4ODg8PDxAQEBERERISEhMTExQUFBUVFRYWFhcXFxgYGBkZ
      GRoaGhsbGxwcHB0dHR4eHh8fHyAgICEhISIiIiMjIyQkJCUlJSYmJicnJygo
      KCkpKSoqKisrKywsLC0tLS4uLi8vLzAwMDExMTIyMjMzMzQ0NDU1NTY2Njc3
      Nzg4ODk5OTo6Ojs7Ozw8PD09PT4+Pj8/P0BAQEFBQUJCQkNDQ0REREVFRUZG
      RkdHR0hISElJSUpKSktLS0xMTE1NTU5OTk9PT1BQUFFRUVJSUlNTU1RUVFVV
      VVZWVldXV1hYWFlZWVpaWltbW1xcXF1dXV5eXl9fX2BgYGFhYWJiYmNjY2Rk
      ZGVlZWZmZmdnZ2hoaGlpaWpqamtra2xsbG1tbW5ubm9vb3BwcHFxcXJycnNz
      c3R0dHV1dXZ2dnd3d3h4eHl5eXp6ent7e3x8fH19fX5+fn9/f4CAgIGBgYKC
      goODg4SEhIWFhYaGhoeHh4iIiImJiYqKiouLi4yMjI2NjY6Ojo+Pj5CQkJGR
      kZKSkpOTk5SUlJWVlZaWlpeXl5iYmJmZmZqampubm5ycnJ2dnZ6enp+fn6Cg
      oKGhoaKioqOjo6SkpKWlpaampqenp6ioqKmpqaqqqqurq6ysrK2tra6urq+v
      r7CwsLGxsbKysrOzs7S0tLW1tba2tre3t7i4uLm5ubq6uru7u7y8vL29vb6+
      vr+/v8DAwMHBwcLCwsPDw8TExMXFxcbGxsfHx8jIyMnJycrKysvLy8zMzM3N
      zc7Ozs/Pz9DQ0NHR0dLS0tPT09TU1NXV1dbW1tfX19jY2NnZ2dra2tvb29zc
      3N3d3d7e3t/f3+Dg4OHh4eLi4uPj4+Tk5OXl5ebm5ufn5+jo6Onp6erq6uvr
      6+zs7O3t7e7u7u/v7/Dw8PHx8fLy8vPz8/T09PX19fb29vf39/j4+Pn5+fr6
      +vv7+/z8/P39/f7+/v///yH5BAEAAP8ALAAAAAALAAcAQAgmAIEJHEgQ2L+D
      AgEIPIgQGICHDxcyRDjxX0KIECVeVGiw4sCJAQEAOw==
   }
   # Right arrow
   image create photo Tree::closedmac -data {
      R0lGODlhBwALAPcAAAAAAAEBAQICAgMDAwQEBAUFBQYGBgcHBwgICAkJCQoK
      CgsLCwwMDA0NDQ4ODg8PDxAQEBERERISEhMTExQUFBUVFRYWFhcXFxgYGBkZ
      GRoaGhsbGxwcHB0dHR4eHh8fHyAgICEhISIiIiMjIyQkJCUlJSYmJicnJygo
      KCkpKSoqKisrKywsLC0tLS4uLi8vLzAwMDExMTIyMjMzMzQ0NDU1NTY2Njc3
      Nzg4ODk5OTo6Ojs7Ozw8PD09PT4+Pj8/P0BAQEFBQUJCQkNDQ0REREVFRUZG
      RkdHR0hISElJSUpKSktLS0xMTE1NTU5OTk9PT1BQUFFRUVJSUlNTU1RUVFVV
      VVZWVldXV1hYWFlZWVpaWltbW1xcXF1dXV5eXl9fX2BgYGFhYWJiYmNjY2Rk
      ZGVlZWZmZmdnZ2hoaGlpaWpqamtra2xsbG1tbW5ubm9vb3BwcHFxcXJycnNz
      c3R0dHV1dXZ2dnd3d3h4eHl5eXp6ent7e3x8fH19fX5+fn9/f4CAgIGBgYKC
      goODg4SEhIWFhYaGhoeHh4iIiImJiYqKiouLi4yMjI2NjY6Ojo+Pj5CQkJGR
      kZKSkpOTk5SUlJWVlZaWlpeXl5iYmJmZmZqampubm5ycnJ2dnZ6enp+fn6Cg
      oKGhoaKioqOjo6SkpKWlpaampqenp6ioqKmpqaqqqqurq6ysrK2tra6urq+v
      r7CwsLGxsbKysrOzs7S0tLW1tba2tre3t7i4uLm5ubq6uru7u7y8vL29vb6+
      vr+/v8DAwMHBwcLCwsPDw8TExMXFxcbGxsfHx8jIyMnJycrKysvLy8zMzM3N
      zc7Ozs/Pz9DQ0NHR0dLS0tPT09TU1NXV1dbW1tfX19jY2NnZ2dra2tvb29zc
      3N3d3d7e3t/f3+Dg4OHh4eLi4uPj4+Tk5OXl5ebm5ufn5+jo6Onp6erq6uvr
      6+zs7O3t7e7u7u/v7/Dw8PHx8fLy8vPz8/T09PX19fb29vf39/j4+Pn5+fr6
      +vv7+/z8/P39/f7+/v///yH5BAEAAP8ALAAAAAAHAAsAQAgmAIEB+0eQIDAA
      AgseBIBw4MGEBhk2/AdRYUWKDCEuzCiQo8GLAQEAOw==
   }
   # Selected checkbox
   image create photo Tree::selectbm -data {
      R0lGODdhDQANAPcAAAAAAIAAAACAAICAAAAAgIAAgACAgMDAwMDcwKbK8EAg
      AGAgAIAgAKAgAMAgAOAgAABAACBAAEBAAGBAAIBAAKBAAMBAAOBAAABgACBg
      AEBgAGBgAIBgAKBgAMBgAOBgAACAACCAAECAAGCAAICAAKCAAMCAAOCAAACg
      ACCgAECgAGCgAICgAKCgAMCgAOCgAADAACDAAEDAAGDAAIDAAKDAAMDAAODA
      AADgACDgAEDgAGDgAIDgAKDgAMDgAODgAAAAQCAAQEAAQGAAQIAAQKAAQMAA
      QOAAQAAgQCAgQEAgQGAgQIAgQKAgQMAgQOAgQABAQCBAQEBAQGBAQIBAQKBA
      QMBAQOBAQABgQCBgQEBgQGBgQIBgQKBgQMBgQOBgQACAQCCAQECAQGCAQICA
      QKCAQMCAQOCAQACgQCCgQECgQGCgQICgQKCgQMCgQOCgQADAQCDAQEDAQGDA
      QIDAQKDAQMDAQODAQADgQCDgQEDgQGDgQIDgQKDgQMDgQODgQAAAgCAAgEAA
      gGAAgIAAgKAAgMAAgOAAgAAggCAggEAggGAggIAggKAggMAggOAggABAgCBA
      gEBAgGBAgIBAgKBAgMBAgOBAgABggCBggEBggGBggIBggKBggMBggOBggACA
      gCCAgECAgGCAgICAgKCAgMCAgOCAgACggCCggECggGCggICggKCggMCggOCg
      gADAgCDAgEDAgGDAgIDAgKDAgMDAgODAgADggCDggEDggGDggIDggKDggMDg
      gODggAAAwCAAwEAAwGAAwIAAwKAAwMAAwOAAwAAgwCAgwEAgwGAgwIAgwKAg
      wMAgwOAgwABAwCBAwEBAwGBAwIBAwKBAwMBAwOBAwABgwCBgwEBgwGBgwIBg
      wKBgwMBgwOBgwACAwCCAwECAwGCAwICAwKCAwMCAwOCAwACgwCCgwECgwGCg
      wICgwKCgwMCgwOCgwADAwCDAwEDAwGDAwIDAwKDAwP/78KCgpICAgP8AAAD/
      AP//AAAA//8A/wD//////yH5BAAAAAAALAAAAAANAA0AAAg8AEkJHEiQ4L+D
      CBP+E6iwIcOG/wAcfKgQgMSFpA5ejGhxYkaOHDc+tEgyIcWSJj9qdKgSIkaX
      CAvKHBgQADs=
   }
   # De-Selected checkbox
   image create photo Tree::unselectbm -data {
      R0lGODdhDQANAPcAAAAAAIAAAACAAICAAAAAgIAAgACAgMDAwMDcwKbK8EAg
      AGAgAIAgAKAgAMAgAOAgAABAACBAAEBAAGBAAIBAAKBAAMBAAOBAAABgACBg
      AEBgAGBgAIBgAKBgAMBgAOBgAACAACCAAECAAGCAAICAAKCAAMCAAOCAAACg
      ACCgAECgAGCgAICgAKCgAMCgAOCgAADAACDAAEDAAGDAAIDAAKDAAMDAAODA
      AADgACDgAEDgAGDgAIDgAKDgAMDgAODgAAAAQCAAQEAAQGAAQIAAQKAAQMAA
      QOAAQAAgQCAgQEAgQGAgQIAgQKAgQMAgQOAgQABAQCBAQEBAQGBAQIBAQKBA
      QMBAQOBAQABgQCBgQEBgQGBgQIBgQKBgQMBgQOBgQACAQCCAQECAQGCAQICA
      QKCAQMCAQOCAQACgQCCgQECgQGCgQICgQKCgQMCgQOCgQADAQCDAQEDAQGDA
      QIDAQKDAQMDAQODAQADgQCDgQEDgQGDgQIDgQKDgQMDgQODgQAAAgCAAgEAA
      gGAAgIAAgKAAgMAAgOAAgAAggCAggEAggGAggIAggKAggMAggOAggABAgCBA
      gEBAgGBAgIBAgKBAgMBAgOBAgABggCBggEBggGBggIBggKBggMBggOBggACA
      gCCAgECAgGCAgICAgKCAgMCAgOCAgACggCCggECggGCggICggKCggMCggOCg
      gADAgCDAgEDAgGDAgIDAgKDAgMDAgODAgADggCDggEDggGDggIDggKDggMDg
      gODggAAAwCAAwEAAwGAAwIAAwKAAwMAAwOAAwAAgwCAgwEAgwGAgwIAgwKAg
      wMAgwOAgwABAwCBAwEBAwGBAwIBAwKBAwMBAwOBAwABgwCBgwEBgwGBgwIBg
      wKBgwMBgwOBgwACAwCCAwECAwGCAwICAwKCAwMCAwOCAwACgwCCgwECgwGCg
      wICgwKCgwMCgwOCgwADAwCDAwEDAwGDAwIDAwKDAwP/78KCgpICAgP8AAAD/
      AP//AAAA//8A/wD//////yH5BAAAAAAALAAAAAANAA0AAAgtAEkJHEiQ4L+D
      CBP+E6iwIcOGCR9CPChxYkWIFx2Smogwo0KPETdyXFiwpMCAADs=
   }
   # Partial Selection checkbox
   image create photo Tree::parselectbm -data {
      R0lGODdhDQANAPcAAAAAAIAAAACAAICAAAAAgIAAgACAgMDAwMDcwKbK8EAg
      AGAgAIAgAKAgAMAgAOAgAABAACBAAEBAAGBAAIBAAKBAAMBAAOBAAABgACBg
      AEBgAGBgAIBgAKBgAMBgAOBgAACAACCAAECAAGCAAICAAKCAAMCAAOCAAACg
      ACCgAECgAGCgAICgAKCgAMCgAOCgAADAACDAAEDAAGDAAIDAAKDAAMDAAODA
      AADgACDgAEDgAGDgAIDgAKDgAMDgAODgAAAAQCAAQEAAQGAAQIAAQKAAQMAA
      QOAAQAAgQCAgQEAgQGAgQIAgQKAgQMAgQOAgQABAQCBAQEBAQGBAQIBAQKBA
      QMBAQOBAQABgQCBgQEBgQGBgQIBgQKBgQMBgQOBgQACAQCCAQECAQGCAQICA
      QKCAQMCAQOCAQACgQCCgQECgQGCgQICgQKCgQMCgQOCgQADAQCDAQEDAQGDA
      QIDAQKDAQMDAQODAQADgQCDgQEDgQGDgQIDgQKDgQMDgQODgQAAAgCAAgEAA
      gGAAgIAAgKAAgMAAgOAAgAAggCAggEAggGAggIAggKAggMAggOAggABAgCBA
      gEBAgGBAgIBAgKBAgMBAgOBAgABggCBggEBggGBggIBggKBggMBggOBggACA
      gCCAgECAgGCAgICAgKCAgMCAgOCAgACggCCggECggGCggICggKCggMCggOCg
      gADAgCDAgEDAgGDAgIDAgKDAgMDAgODAgADggCDggEDggGDggIDggKDggMDg
      gODggAAAwCAAwEAAwGAAwIAAwKAAwMAAwOAAwAAgwCAgwEAgwGAgwIAgwKAg
      wMAgwOAgwABAwCBAwEBAwGBAwIBAwKBAwMBAwOBAwABgwCBgwEBgwGBgwIBg
      wKBgwMBgwOBgwACAwCCAwECAwGCAwICAwKCAwMCAwOCAwACgwCCgwECgwGCg
      wICgwKCgwMCgwOCgwADAwCDAwEDAwGDAwIDAwKDAwP/78KCgpICAgP8AAAD/
      AP//AAAA//8A/wD//////yH5BAAAAAAALAAAAAANAA0AAAg8APMJHEiQ4L+D
      CBP+E6iwIcOG/w4cfKjwgMSF+Q5ejGhxYkaOHDc+tEgyIcWSJj9qdKgSIkaX
      CAvKHBgQADs=
   }
   #
   # This defines the "File" image (a picture of a text file).
   #
   image create photo Tree::ifile -data {
       R0lGODdhEAAQAPIAAAAAAHh4eLi4uPj4+P///wAAAAAAAAAAACwAAAAAEAAQAAADPkixzPOD
       yADrWE8qC8WN0+BZAmBq1GMOqwigXFXCrGk/cxjjr27fLtout6n9eMIYMTXsFZsogXRKJf6u
       P0kCADv/
   }

   # Changes the checkbox
   variable TreeImages
   array set TreeImages {
      1     Tree::selectbm
      true  Tree::selectbm
      True  Tree::selectbm
      0     Tree::unselectbm
      false Tree::unselectbm
      False Tree::unselectbm
      2     Tree::parselectbm
   }
   # Set by -treeLine
   variable OpenCloseImages
   array set OpenCloseImages {
      0,open Tree::openmac
      1,open Tree::openbm
      0,closed Tree::closedmac
      1,closed Tree::closedbm
   }

}
##
# Return whether a user clicked a checkbox or
# someone externally set the dirty flag
# Used to flag when to save a configuration file
#
proc Tree::isDirty {w} {
   upvar #0 $w myTree
   return $myTree(-isDirty)
}
##
# Flag the tree dirty from an external source
#
proc Tree::setDirty {w {how 1}} {
   upvar #0 $w myTree
   set myTree(-isDirty) $how
   return
}

#
# Create a new tree widget.  $args become the configuration arguments to
# the canvas widget from which the tree is constructed.
# Don't create the canvas first; it will be created in this procedure.
# Example:
#   frame .f1
#   Tree:create .f1.myCanvas -foreground $fg -background $bg (...other canvas options)
#
proc Tree::create {w args} {
   upvar #0 $w myTree
   set myTree(-isDirty) 0
   set myTree(-rippleSet) 0
   set myTree(-treeLine) 0
   eval canvas $w $args
   bind $w <Destroy> [list Tree::delitem $w /]
   # cb to change outline
   bind [winfo parent $w] <FocusIn>  [list Tree::FocusChange $w 1]
   bind [winfo parent $w] <FocusOut> [list Tree::FocusChange $w 0]
   Tree::dfltconfig $w /
   Tree::buildwhenidle $w
   set myTree(selection) {}
   set myTree(selidx) {}
   bind $w <Configure> [list Tree::canvasResize %W %w %h]
}

##
# Called on a canvas resize or configure event, need to
# The canvas size and the scroll region
proc Tree::canvasResize {win W H} {
   set bbox [$win bbox all]
   $win configure -scrollregion $bbox -yscrollincrement 0.1i
}
# Initialize a element of the tree.
# Internal use only
#
proc Tree::dfltconfig {w v} {
   upvar #0 $w myTree
   set myTree($v:children) {}
   set myTree($v:open) 0
   set myTree($v:icon) ""        ;# Default the icon to 'no icon'
   set myTree($v:tags) {}
}

#
# Pass configuration options to the tree widget
#
proc Tree::widgetConfig {w args} {
   eval $w config $args
}
##
#  Set some value for this tree
#
proc Tree::configure {w setting value} {
   upvar #0 $w myTree
   if {[info exists myTree($setting)]} {
      set myTree($setting) $value
   } else {
      return -code error "Setting $setting does not exist"
   }
}
##
# Return the value of some element in the tree
#
proc Tree::cget {w setting} {
   upvar #0 $w myTree
   if {[info exists myTree($setting)]} {
      return $myTree($setting)
   } else {
      return -code error "Setting $setting does not exist"
   }
}
#----------------------------------------------------------------------------
# Return whether a node exists or not in the tree
#----------------------------------------------------------------------------
proc Tree::exists {w v args} {
   upvar #0 $w myTree
   # If windows path or unc, strip drive letters / double slashes to make a Unix style root
   set v [string map {// / : "_Drive"} $v]

   # Force this to be off the root in case user forget leading /
   if {[string index $v 0] != "/"} {
      set v "/$v"
   }
   set dir [file dirname $v]
   set n [file tail $v]
   # puts "$v - $dir - $n"
   # If this is a root
   set i [lsearch -sorted -exact $myTree($dir:children) $n]
   if {$i>=0} {
      return 1
   } else {
      return 0
   }
}
#----------------------------------------------------------------------------
# Determine if a node as a parent
#----------------------------------------------------------------------------
proc Tree::hasParent {w v} {
   upvar #0 $w myTree
   # If windows path or unc, strip drive letters / double slashes to make a Unix style root
   set v [string map {// / : "_Drive"} $v]

   # Force this to be off the root in case user forget leading /
   if {[string index $v 0] ne "/"} {
      set v "/$v"
   }
   set parent [file dirname $v]
   return [info exists myTree($parent:children)]
}

#
# Insert a new element $v into the tree $w.
# Example:
#   Tree:newitem .f1.myCanvas -image Tree:ROM (...other node options)
#
# args:
#    -image - the icon associated with a node.  This is typically an open or
#        closed file folder.  If you don't specify an icon then no icon will
#        be displayed; this looks like Richard Hipp's original version.
#    -tags - any canvas tags that you wish to associate with this node
#    -variable - a variable name that you want to reflect the nodestate
#        of this node.  You must specify this if you want the node to get a
#        selection box (a checkbox).  If you specify this option then you
#        don't have to do anything else; the node will automatically get a
#        checkbox.  (You must create the variable first.)
#
# The '+' and '-' boxes that open and close subdirectories are created
# automatically in Tree:buildlayer.
#
proc Tree::newitem {w v args} {
   upvar #0 $w myTree
   # If windows path or unc, strip drive letters / double slashes to make a Unix style root
   set v [string map {// / : "_Drive"} $v]

   # Force this to be off the root in case user forget leading /
   if {[string index $v 0] != "/"} {
      set v "/$v"
   }
   set dir [file dirname $v]
   set n [file tail $v]
   # puts "$v - $dir - $n"
   # Ensure parent exists, if not bail
   if {![info exists myTree($dir:children)]} {
      #eval Tree::newitem $w $dir $args
      return 0
   }
   # Get index of node, if it exists just return
   set i [lsearch -sorted -exact $myTree($dir:children) $n]
   if {$i >= 0} {
      return 1
   }
   lappend myTree($dir:children) $n
   set myTree($dir:children) [lsort $myTree($dir:children)]
   Tree::dfltconfig $w $v
   set myTree($v:parent) $dir ;#LB added
   foreach {op arg} $args {
      switch -exact -- $op {
         -image {set myTree($v:icon) $arg}
         -tags {set myTree($v:tags) $arg}
         -variable {
            upvar #0 $arg myVar
            set myTree($v:variable) $arg
         }
      }
   }
   Tree::buildwhenidle $w
   return 1
}
#
# Delete element $v from the tree $w.  If $v is /, then destroy the entire widget.
# In all cases, go through the 'Tree' array and remove all attributes of this node
# and its children.
#
proc Tree::delitem {w v} {
   upvar #0 $w myTree
   if {![info exists myTree($v:open)]} return
   #
   # Are we deleting the entire tree?
   #
   if {[string equal $v /]} {
      # delete the whole widget
      destroy $w
      unset -nocomplain myTree
      return
   }
   #
   # We are not deleting the entire tree, so first delete the children
   # of this node.
   #
#   foreach c $myTree($v:children) {
#      if {[catch {Tree::delitem $w [file join $v $c]}] } then {
#         unset -nocomplain myTree($c)
#      }
#   }
   #remove tag:240
   #(/D Media/Madman Across The Water:tag)         = 240

   #
   # Now delete all array elements referring to this node.
   #
   if {[info exists myTree($v:parent)]} {
      set dir $myTree($v:parent)
      #set dir [file dirname $v]
      if {[info exists myTree($dir:children)]} {
         set n [file tail $v]
         set i [lsearch -sorted -exact $myTree($dir:children) $n]
         if {$i>=0} {
            # remove this entry
            set myTree($dir:children) [lreplace $myTree($dir:children) $i $i]
         }
      }
   }
   # remove all tags to this node
   foreach {index tag} [array get myTree $v:tags] {
      unset -nocomplain myTree(tag:$tag)
   }
   # Remove this nodes data
   array unset myTree $v:*
   array unset myTree $v/*:*
   Tree::buildwhenidle $w
}

#
# Change the selection to the indicated item
#
proc Tree::setselection {w v} {
   upvar #0 $w myTree
   set myTree(selection) $v
   Tree::drawselection $w
}

#
# Retrieve the current selection
#
proc Tree::getselection w {
   upvar #0 $w myTree
   return $myTree(selection)
}

#
# Retrieve the current selection tag
#
proc Tree::activetag w {
   upvar #0 $w myTree
   return $myTree(activeTag)
}
# Internal use only.
# Draw the tree on the canvas
proc Tree::build {w} {
   upvar #0 $w myTree
   $w delete all
   # Remove each of the old tag entries
   array unset myTree tag:*
   #foreach tag [array names myTree tag:*] {
   #   unset myTree($tag)
   #}
   set myTree(y) 30
   Tree::buildlayer $w / 10   ;# This builds the root
   $w config -scrollregion [$w bbox all]
   Tree::drawselection $w
}

# Internal use only.
# Build a single layer of the tree on the canvas.  Indent by $in pixels
proc Tree::buildlayer {w v in} {
   upvar #0 $w myTree
   variable TreeImages
   variable font
   variable treeLine
   variable indent
   variable OpenCloseImages
   # puts "Tree::buildlayer $w $v $in"
   set deltaHeight [font metrics $font -linespace]  ;# Node spacing is line spacing
   set deltaHeight [expr {$deltaHeight < 15?15:$deltaHeight}]  ;# Clamp to min size
   set y $myTree(y)
   set start [expr {$y-10}]
   set j 0
   set tl $myTree(-treeLine)
   foreach c $myTree($v:children) {
      set y $myTree(y)
      incr myTree(y) $deltaHeight ;#17
      set token [file join $v $c]
      # token is the folder name
      set taglist x
      # BUG: ensure $token:tags exists (it should always exist)
      if {[info exists myTree($token:tags)]} {
         foreach tag $myTree($token:tags) {
            lappend taglist $tag
         }
      } else {
         debug "Missing:$token:tags"
      }
      set x [expr {$in+10}]  ;# how far over to place the icon from the +/-
      ##
      # Create the user specified icon.  Use '$icon.close' if that exists; if
      # it doesn't exist then use whatever the user specified.
      #
      if {[string length $myTree($token:icon)] > 0} {
         set icon $myTree($token:icon)
         if {![catch {image inuse $icon.close}] } then {
            append icon ".close"
         }
         set k [$w create image $x $y -image $icon -anchor w -tags $taglist]
         set myTree(tag:$k) $token
         # Create the horizontal line that joins the icon to the vertical line
         # (going to this node's parent).
         if {$tl} {
            $w create line $in $y [expr {$in+12}] $y -fill gray ;#50
         }
      } else {
         # The user did not specify an icon, so draw a longer line
         # TFW er, how long should this line really be? Idealy the
         # the width of an icon had it been there, whatever that is
         if {$tl} {
            $w create line $in $y [expr {$in+29}] $y -fill gray ;#50
         }
      }
      ##
      # If an icon was specified it should be ~16 pixels wide
      # the next incr should accomodate this width
      incr x 20
      # Toggle button if user wants it (linked to a variable)
      if {[info exists myTree($token:variable)]} {
         upvar #0 $myTree($token:variable) myVar
         if {![info exists myVar]} {
            # default to image 0
            set myVar 0
         }
         incr x 4 ;# Spacing between user icon and checkbox
         set cb [$w create image $x $y -image $TreeImages($myVar)]
         # On a click, need to change indicator and set a global
         $w bind $cb <1> [list Tree::toggleNode $w $cb $token]
         # Move right for the text
         incr x [expr {[image width $TreeImages($myVar)]-3}]
      }
      # Place the node text
      set j [$w create text $x $y -text $c -font $font -fill $Tree::fg -activefill $Tree::fg \
         -anchor w -tags $taglist]
      set myTree(tag:$j) $token
      set myTree($token:tag) $j
      # Create the +/- images to click on to expand the tree
      if {[string length $myTree($token:children)]} {
         if {$myTree($token:open)} {
            # - size box
            set j [$w create image $in $y -image $OpenCloseImages($tl,open)]
            $w bind $j <1> [list Tree::expCallBack $w $token 0]
            Tree::buildlayer $w $token [expr {$in+$indent}]
         } else {
            # + sign box
            set j [$w create image $in $y -image $OpenCloseImages($tl,closed)]
            $w bind $j <1> [list Tree::expCallBack $w $token 1]
         }
      }
   }
   if {$tl} {
      set j [$w create line $in $start $in [expr {$y+1}] -fill gray]
   }
   $w lower $j
}
#------------------------------------------------------------------------------
# Function   : Tree::expCallBack
# Description: Respond to clicking the +/- expand/contract buttons
# Author     : Tom Wilkason
# Date       : 7/30/2003
#------------------------------------------------------------------------------
proc Tree::expCallBack {w token val} {
   upvar #0 $w myTree
   set myTree($token:open) $val
   Tree::build $w
}
#------------------------------------------------------------------------------
# Function   : Tree::toggleNode
# Description: Bound to checkbox at node
# Author     : Tom Wilkason
# Date       : 1/9/2001
#------------------------------------------------------------------------------
proc Tree::toggleNode {w tag node} {
   upvar #0 $w myTree
   upvar $myTree($node:variable) linkedVariable
   # If the nodeState=2 (Partially Selected) then change it to 1 (Selected).
   # After this, clicking on the checkbox will toggle the node between
   # Selected and Not Selected.
   if {$linkedVariable==2} {
      set linkedVariable 0
   }
   # Toggle the nodeState
   set linkedVariable [expr {!$linkedVariable}]
   # Select the correct Icon for the new state.
   $w itemconfigure $tag -image $Tree::TreeImages($linkedVariable)
   # Flag the dirty bit
   set myTree(-isDirty) 1
   ##
   # Special behavior based on settings, ripple the changes up and down
   # the tree.
   #
   if {$myTree(-rippleSet)} {
      #
      # If this node is not a leaf, then set everything below it to the same
      # nodestate.
      #
      Tree::setChildState $w $node 1
      #
      # Now set all of this nodes' parents to their correct state.
      #
      Tree::setParentState $w $node $linkedVariable
      Tree::buildwhenidle $w
   }
}
##
# Sets the state of child nodes to the state of the parent.  If 'recurse'
# is non-zero, it sets ALL children to the state of the parent; otherwise
# it sets just the state of the immediate children.  'Recurse' defaults to
# 'True'.
#
proc Tree::setChildState {w v {recurse 1}} {
   upvar #0 $w myTree
   upvar #0 $myTree($v:variable) parentVariable
   foreach child $myTree($v:children) {
      #
      # Set this node's linked variable to the nodestate.  The user could
      # create a badly constructed tree that has parent nodes with checkboxes,
      # but child nodes without.  If the user does this, then this code will
      # throw an error, but I think that is OK.
      #
      upvar #0 $myTree($v/$child:variable) linkedVariable
      set linkedVariable $parentVariable
   }
   if {$recurse} {
      foreach child $myTree($v:children) {
         Tree::setChildState $w $v/$child 1
      }
   }
}



#------------------------------------------------------------------------------
#
# proc Tree::setParentState {w v childState}
#
# Sets the state of parent nodes based on the state of the children.
# The states are:
# 0 - Not Selected       - (empty box)
# 1 - Selected           - (checkmarked box)
# 2 - Partially Selected - (grey checkmark in a red box)
#
# The State-Change rules are:
#
# - 1) If the child node is set to '0', then the parent state cannot
#      become '1' - it must be either '0' or 'Partial'.  Therefore, start
#      testing the states of the parent's children.  If any child state is
#      1 or 2, then the parent state must be 'Partial'.  Set it to 'Partial'.
#   2) If we have evaluated all of the children and all are '0', then set the
#      parent state to '0'.
#   3) The parent becomes the new child node; recurse.
#
# - 1) If the child node is set to '1', then the parent state cannot
#      become '0' - it must be either '1' or 'Partial'.  Therefore, start
#      testing the states of the parent's children.  If any child state is
#      0 or 2, then the parent state must be 'Partial'.  Set it to 'Partial'.
#   2) If we have evaluated all of the children and all are '1', then set the
#      parent state to '1'.
#   3) The parent becomes the new child node; recurse.
#
# - If the child node is set to 'Partial', then set its parent state to
#   'Partial'.
#
# There are simpler algorithms than this one, but this one is efficient - it
# terminates as soon as it knows the state, rather than exhaustively testing
# every child node.  I am going to make the further simplifying assumption
# that the tree was built correctly to begin with; this means that we don't
# have to test sub-children of a node; testing the direct children is
# sufficient.  If the user builds an incorrect tree (initializes it
# incorrectly) then this will fail.  Think of this diagram:
#
#            Node A (1)
#              |
#       --------------
#       |             |
#     Node B (1)    Node C (2)
#                     |
#                   Node D (0)
#
# Notice that Node C does not correctly reflect the state of it's children -
# it has only one child, with a state of zero, so Node C should also be zero.
# If Node B changes to zero, then Node A should also be zero, but using an
# abbreviated tree evaluation (rather than a full sub-tree evaluation) will
# cause Node A to get a value of '2'.
#
#------------------------------------------------------------------------------
proc Tree::setParentState {w v childState} {
   upvar #0 $w myTree
   set parent $myTree($v:parent)
   #
   # When we reach the root then we are done.
   #
   if {[string equal $parent "/"]} {
      return 0
   }
   upvar #0 $myTree($parent:variable) parentVariable
   #
   # Case 1 - child state == 0 - 'Not Selected'
   #
   switch -- $childState {
      0 {
         set evalState [Tree::evalChildren $w $parent 1 2]
         if {$evalState ==1 || $evalState ==2} {
            set parentVariable 2
         } else {
            set parentVariable 0
         }
      }
      1 {
         set evalState [Tree::evalChildren $w $parent 0 2]
         if {$evalState ==0 || $evalState ==2} {
            set parentVariable 2
         } else {
            set parentVariable 1
         }
      }
      2 {
         set parentVariable 2
      }
   }
   #
   # Continue up the tree.
   #
   Tree::setParentState $w $parent $parentVariable
   return 1
}

#------------------------------------------------------------------------------
#
# proc Tree::evalChildren {w v state1 state2}
#
# Searches the current node's children for a node with one of the two
# specified states.  It returns the first matching state it finds.  It
# searches only the immediate children; not sub-children.  If there are no
# children (a leaf node) then it compares the state of the current node to
# the desired states, and returns the matched value if one matches.  If all
# of these tests fail it returns -1.
#
#------------------------------------------------------------------------------
proc Tree::evalChildren {w v state1 state2} {
   upvar #0 $w myTree
   if {[llength $myTree($v:children)] != 0} {
      foreach child $myTree($v:children) {
         upvar #0 $myTree($v/$child:variable) linkedVariable
         set state $linkedVariable
         if {$state==$state1 || $state==$state2} {
            return $state
         }
      }
      return -1
   } else {    ;# This is a leaf node
      upvar #0 $myTree($v:variable) linkedVariable
      if {$linkedVariable==$state1 || $linkedVariable==$state2} {
         return $linkedVariable
      } else {
         return -1
      }
   }
}

#------------------------------------------------------------------------------
#
# proc Tree::nodes {w v state}
#
# Searches the tree w, starting at node v, and returns a list of all children
# with nodeState 'state'.  'State' defaults to '1' (selected or checked).
#
#------------------------------------------------------------------------------
proc Tree::nodes {w v {state 1}} {
   upvar #0 $w myTree
   set nodes {}
   foreach c $myTree($v:children) {
      set token [file join $v $c]
      # token is the folder name
      # If a linked variable exists, and it has the desired value, then append
      # this nodes' name to the list.
      if {[info exists myTree($token:variable)]} {
         upvar #0 $myTree($token:variable) myVar
         if {$myVar == $state} {lappend nodes $token}
      }
      # If this node has children, then recurse into the children
      if {[string length $myTree($token:children)]} {
         # Append each child node to the list of nodes (don't append the list,
         # because then we would have a list of lists, not a list of nodes.)
         # We don't have to test the state, as that has already been done.
         foreach node [Tree::nodes $w $token $state] {
            lappend nodes $node
         }
      }
   }
   return $nodes
}

#------------------------------------------------------------------------------
# Return true if the node is open
#------------------------------------------------------------------------------
proc Tree::isOpen {w v} {
   upvar #0 $w myTree
   if {[info exists myTree($v:open)] && $myTree($v:open)==1} {
      return 1
   } else {
      return 0
   }
}
##
# Toggles between expanded and not, can bind to a key
#
proc Tree::toggleExpand {w v {recurse 0}} {
   upvar #0 $w myTree
   if {[info exists myTree($v:open)] && $myTree($v:open)==1} {
      Tree::close $w $v $recurse
   } else {
      Tree::open  $w $v $recurse
   }
}
# Open a branch of a tree from this point forward
# Added: Make sure branch is also opened from the root forward
#
proc Tree::open {w v {recurse 0}} {
   upvar #0 $w myTree
   # Open to root first
   set parent $v
   while {[info exists myTree($parent:parent)]} {
      set parent $myTree($parent:parent)
      set myTree($parent:open) 1
   }
   # Then open up required levels
   Tree::_open $w $v $recurse
}
proc Tree::_open {w v {recurse 0}} {
   upvar #0 $w myTree
   # puts "Tree::open $w $v"  && $myTree($v:open)==0
   # Open to root also  && $myTree($v:open)==0
   if {[info exists myTree($v:open)]
      && [info exists myTree($v:children)]
      && [llength $myTree($v:children)] > 0} {
      set myTree($v:open) 1
      Tree::buildwhenidle $w
      # Recurse to a certain depth
      if {$recurse >= [llength [file split $v]]} {
         foreach c $myTree($v:children) {
            Tree::_open $w [file join $v $c] $recurse
         }
      }
   }
}
proc Tree::close {w v {recurse 0}} {
   upvar #0 $w myTree
   if {[info exists myTree($v:open)] && $myTree($v:open)==1} {
      set myTree($v:open) 0
      Tree::buildwhenidle $w
      if {$recurse} {
         foreach c $myTree($v:children) {
            Tree::close $w [file join $v $c] $recurse
         }
      }
   }
}
# Internal use only.
# Change highlight to dashed border when focus changes
proc Tree::FocusChange {w In} {
   upvar #0 $w myTree
   if {$In} {
      $w itemconfigure $myTree(selidx) -state normal -outline {}
   } else {
      $w itemconfigure $myTree(selidx) -state disabled -outline black
   }
}
# Draw the selection highlight
proc Tree::drawselection {w} {
   variable Tree
   upvar #0 $w myTree
   if {[string length $myTree(selidx)]} {
      $w delete $myTree(selidx)
   }
   set v $myTree(selection)
   if {[string length $v]==0} return
   if {![info exists myTree($v:tag)]} return
   set bbox [$w bbox $myTree($v:tag)]
   set myTree(activeTag) $myTree($v:tag)
   if {[llength $bbox]==4} {
      set i [eval $w create rectangle $bbox \
         -outline $Tree::fg \
         -disabledoutline $Tree::fg\
         -fill $Tree::highlight \
         -disabledfill $Tree::bg \
         -dash 1]
      set myTree(selidx) $i
      $w lower $i
   } else {
      set myTree(selidx) {}
   }
   ##
   # Change the open/close folder tag
   #
   # Hack: Folder icon tag is two below text tag (1 below is checkbox)
   catch {
      set tag [expr {$myTree($v:tag)-2}]
      set folder $myTree(tag:$tag)
      set icon $myTree($folder:icon)
      ##
      # Configure the icon as open or closed as appropriate
      #
      if {[info exists myTree(lastFI)]} {
         $w itemconfigure $myTree(lastFI) -image $myTree(lastIcon).close
      }
      $w itemconfigure $tag -image $icon.open
      set myTree(lastFI) $tag
      set myTree(lastIcon) $icon
   }
}

# Internal use only
# Call Tree::build then next time we're idle
proc Tree::buildwhenidle {w} {
   upvar #0 $w myTree
   after cancel [list Tree::build $w]
   after idle   [list Tree::build $w]
}

#
# Return the tag for widget $w that is located
# at real coordinates $x, $y
#
proc Tree::tagat {w x y} {
   set x [$w canvasx $x]
   set y [$w canvasy $y]
   upvar #0 $w myTree
   foreach m [$w find overlapping $x $y $x $y] {
   #$myTree($v:tag)
      if {[info exists myTree(tag:$m)]} {
         return $m
      }
   }
   return ""
}
#
# Return the full pathname of the label for widget $w that is located
# at real coordinates $x, $y
#
proc Tree::labelat {w x y} {
   set x [$w canvasx $x]
   set y [$w canvasy $y]
   upvar #0 $w myTree
   foreach m [$w find overlapping $x $y $x $y] {
      if {[info exists myTree(tag:$m)]} {
         return $myTree(tag:$m)
      }
   }
   return ""
}
##
# Debugging stuff
#
proc TraceTree {} {
   trace variable ::Tree w DumpTree
}
proc DumpTree {_G I H} {
   upvar $_G G
   puts "$I = $G($I)"
}
package provide saTree 1.0

