#------------------------------------------------------------------------------
# soundControl Package 1.0
#
# Provides interfaces to the snack sound engine
#
# Copyright (C) 2001 Tom Wilkason
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, autoLevelInc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
#
# Please send any comments/bug reports to
# tom.wilkason@cox.net  (Tom Wilkason)
#
#------------------------------------------------------------------------------
set [file rootname [file tail [info script]]]_History {
   $Header: /cvsroot/snackamp/snackamp/lib/soundControl.tcl,v 1.26 2007/10/01 01:37:33 wilkason Exp $
}
#------------------------------------------------------------------------------
# Function   : soundControl namespace
# Description: All snack related calls are in here
# Author     : Tom Wilkason
# Date       : 6/28/2002
#------------------------------------------------------------------------------
namespace eval soundControl {
   variable paused  0
   variable butRelief
   variable butState
   variable startOffset 0.0
   variable playState "Stopped"
   variable Thresh 10   ;# todo: Link to setting (maybe not)
   variable soundObj
   variable infoObj
   variable lastRate 0
   variable lastType file
   variable autoLevelInc
   variable SongPlayMode
   variable normalPlayLatency
   variable longPlayLatency
   variable autoLevelMaxSamples
   variable autoLevelWindow
   variable autoLevelReference
   variable gainFactor
   variable Trace
   variable acGaindB -20
   variable globalGain 1.0
   variable trackGain 1.0
   variable globalMixer
   variable lineSelect
   variable lastGlobalGaindB
   variable Title ""
   upvar #0 snackAmpSettings(lastGlobalGaindB)          lastGlobalGaindB
   upvar #0 snackAmpSettings(globalMixer)             globalMixer
   upvar #0 snackAmpSettings(lineSelect)              lineSelect
   upvar #0 snackAmpSettings(SongPlayMode)            SongPlayMode
   upvar #0 snackAmpSettings(autoLevelInc)            autoLevelInc
   upvar #0 snackAmpSettings(normalPlayLatency)       normalPlayLatency
   upvar #0 snackAmpSettings(longPlayLatency)         longPlayLatency
   upvar #0 snackAmpSettings(autoLevelMaxSamples)     autoLevelMaxSamples
   upvar #0 snackAmpSettings(autoLevelWindow)         autoLevelWindow
   upvar #0 snackAmpSettings(autoLevelReference)      autoLevelReference
   upvar #0 snackAmpSettings(gainFactor)              gainFactor
   upvar #0 snackAmpSettings(Trace)                   Trace
   array set butRelief {0 flat   1 sunken 2 flat}
   array set butState {0 normal 1 normal 2 disabled}
   # Try and load ogg vorbis support
   if {![catch {package require snackogg} result]} {
      lappend extTypes  {OGG .ogg}
      lappend loadTypes {{Ogg Vorbis Files} {.ogg}}
      lappend loadKeys  OGG
      lappend saveTypes {{Ogg Vorbis Files} {.ogg}}
      lappend saveKeys  OGG
      snack::addExtTypes $extTypes
      snack::addLoadTypes $loadTypes $loadKeys
      snack::addSaveTypes $saveTypes $saveKeys
   } else {
      puts stderr $result
   }
   # Create the snack sound objects
   if {[catch {snack::sound} soundObj] } then {
      waitWindow::close
      puts stderr $::errorInfo
      tk_messageBox -icon error \
         -title "Fatal Error" \
         -type ok \
         -message "The Snack Engine failed to create a sound object, check the snack library."
      exit
   }
   ;# Alternate sound object for information, cataloging etc.
   if {[catch {snack::sound} infoObj] } then {
      waitWindow::close
      puts stderr $::errorInfo
      tk_messageBox -icon error \
         -title "Fatal Error" \
         -type ok \
         -message "The Snack Engine failed to create a sound object, check the snack library."
      exit
   }
   if {[llength [snack::mixer lines]] == 0} {
      waitWindow::close
      puts stderr $::errorInfo
      tk_messageBox -icon error \
         -title "Fatal Error" \
         -type ok \
         -message "There are no available mixer lines!"
      exit
   }
}
#------------------------------------------------------------------------------
# Function   : soundControl::Timer
# Description: soundControl::Timer to update the display data, active while file playing
#              Ramp the level down the final two seconds to handle truncated songs.
#------------------------------------------------------------------------------
proc soundControl::Timer {args} {
   after cancel soundControl::Timer ;# make sure this is only one running
   global TA
   # Time is curren pos/total len * total time
   # If user dragging slider, don't also try to set slider
   if {$TA(setdrag)} {
      catch {
         set gui::trackPosition [soundControl::position] ;# scale
         gui::updateTime ;# time counters
      }
   }
   if {$::snackAmpSettings(showMS)} {
      after 100 soundControl::Timer
   } else {
      after 250 soundControl::Timer
   }
}
#------------------------------------------------------------------------------
# Function   : getNextCommand
# Description: Return the approprite play command based on settings
# Author     : Tom Wilkason
# Date       : 4/6/2002
#------------------------------------------------------------------------------
proc soundControl::getNextCommand {{delay 0}} {
   Trace
   variable SongPlayMode
   # See if we want to repeat this song
   switch -- $SongPlayMode {
      0 {set nextCommand [list after $delay Play next]}
      1 {set nextCommand {}}
      2 {set nextCommand [list after $delay Play current]}
      default {}
   }
   return $nextCommand
}
#------------------------------------------------------------------------------
# Function   : getNextCommandBump
# Description: Return the approprite play command based on settings, include
#              code to bump the play count. Only used for auto next commands
# Author     : Tom Wilkason
# Date       : 4/6/2002
#------------------------------------------------------------------------------
proc soundControl::getNextCommandBump {track {delay 0}} {
   Trace
   variable SongPlayMode
   # See if we want to repeat this song
   switch -- $SongPlayMode {
      0 {set nextCommand [list after $delay Play next]\;[list db::bumpPlayed $track]}
      1 {set nextCommand {}}
      2 {set nextCommand [list after $delay Play current]\;[list db::bumpPlayed $track]}
      default {}
   }
   return $nextCommand
}
#------------------------------------------------------------------------------
# Function   : soundControl::latency
# Description: Adjust latency for long operations, normally a big latency
#              makes the cpu usage jump due to the rapid decode that fills
#              the output buffer with audio.
# Author     : Tom Wilkason
# Date       : 10/29/2001
#------------------------------------------------------------------------------
proc soundControl::latency {{how normal}} {
   variable normalPlayLatency
   variable longPlayLatency
   if {$how == "normal"} {
      snack::audio playLatency $normalPlayLatency
   } else {
      snack::audio playLatency $longPlayLatency
   }
   Update
}
#------------------------------------------------------------------------------
# Function   : soundControl::length
# Description: Return sample length of sound object
# Author     : Tom Wilkason
#------------------------------------------------------------------------------
proc soundControl::length {args} {
   variable soundObj
   return [$soundObj length]
}
#------------------------------------------------------------------------------
# Function   : soundControl::sinfo
# Description: Return info for sound object
# Author     : Tom Wilkason
#------------------------------------------------------------------------------
proc soundControl::sinfo {args} {
   variable soundObj
   return [$soundObj info]
}

#------------------------------------------------------------------------------
# determine the bitrate of the current track
#------------------------------------------------------------------------------
proc soundControl::bitrate {args} {
   variable soundObj
   set ext [$soundObj config -fileformat]
   switch -- $ext {
      MP3 {set rate [expr {[$soundObj config -bitrate]/1000}]}
      OGG {set rate [expr {[$soundObj config -nominalbitrate]/1000}]}
      WAV {set rate [expr {[$soundObj config -rate]/1000}]}
      default {set rate 0}
   }
   return $rate
}
#------------------------------------------------------------------------------
# Function   : soundControl::duration
# Description: Return duration of sound object
# Author     : Tom Wilkason
#------------------------------------------------------------------------------
proc soundControl::duration {args} {
   variable soundObj
   return [$soundObj length -units seconds]
}
#------------------------------------------------------------------------------
# Function   : soundControl::reset
# Description: Recreate the sound object
# Author     : Tom Wilkason
#------------------------------------------------------------------------------
proc soundControl::reset {{type def}} {
   variable soundObj
   variable lastType
   ##
   # It appears that a cpu race condition can ensue
   # if the sound object is deleted between tracks
   # that are being played together with no or a
   # very short gap between them (on linux anyway)
   # so we want to be careful when delting the sound object
   if {$lastType ne $type} {
#debug "Resetting sound object $type old $lastType"
      if {[info command $soundObj] ne ""} {
         $soundObj stop
         while {[snack::audio elapsedTime] > 0.0} {
            after 20
            Update
         }
         $soundObj destroy
      }
      set soundObj [snack::sound]
      set lastType $type
   }
}
#------------------------------------------------------------------------------
# Function   : soundControl::trackInfo
# Description: Return info on the song
# Author     : Tom Wilkason
# Date       : 4/23/2002
#------------------------------------------------------------------------------
proc soundControl::trackInfo {_Details fileName} {
   variable infoObj
   upvar $_Details Details
   ##
   # Note: config -file can be slow on a network share, even though
   # reading ID3 tags is very fast. Not sure why
   if {[catch {$infoObj config -file $fileName -fileformat [getType $fileName]} result]} {
      array set Details {
         Rate       0
         sampRate 0.0
         Sample   0.0
         Duration   0
         duration "00:00"
      }
      return 0
   } else {
      # Get the bitrate info
      if {[catch {
         set Details(ext) [$infoObj config -fileformat]
         switch -- $Details(ext) {
            MP3 {set rate [expr {[$infoObj config -bitrate]/1000}]}
            OGG {set rate [expr {[$infoObj config -nominalbitrate]/1000}]}
            WAV {set rate [expr {[$infoObj config -rate]/1000}]}
            default {set rate 0}
         }
         set Details(sampRate) [expr {double([$infoObj cget -frequency])/1000.}]
         set len [$infoObj length -units sec]
      }]} then {
         set rate 0
         set Details(sampRate) 0
         set len 0
      }
      set Details(Rate) $rate
      set Details(Sample) $Details(sampRate)
      set Details(Duration) $len
      set Details(duration) [HMSfrac $len 1]
      $infoObj config -file ""
   }
   return 1
}
#------------------------------------------------------------------------------
# Function   : playState
# Description: Return the play state
# Author     : Tom Wilkason
#------------------------------------------------------------------------------
proc soundControl::playState {args} {
   variable playState
   return $playState
}
#------------------------------------------------------------------------------
# Function   : calibrateAudio
# Description: Command line utility function to calibrate a bunch of songs.
# Author     : Tom Wilkason
# Date       : 11/10/2001
#------------------------------------------------------------------------------
proc soundControl::calibrateAudio {{recompute "yes"}} {
   set done 0
   set total 0
   ##
   # Compute the task
   #
   set todo [list]
   foreach {file} [db::sortedList] {
      if {[gainControl::exists $file]} {
         incr done
      } else {
         lappend todo $file
      }
      incr total
   }
   if {($total-$done) == 0} {
      return
   }
   if {[tk_messageBox \
-message "Are you sure you want to auto-level [commify [expr {$total-$done}]]  tracks?
No track files will be modified.\n\nThis may take a while..." \
-icon question -type yesno] != "yes"} {
      return
   }
   #soundControl::latency long
   waitWindow::show "Checking Existing Files"
   ##
   # Now do the work
   #
   set i 0
   foreach {file} $todo {
      waitWindow::percent $done $total "Auto-Leveling\n[splitFile $file]" 1
      if {[waitWindow::ifbreak]} {
         soundControl::latency
         db::commit
         waitWindow::close
         return
      }
      if {$recompute == "no" || ![gainControl::exists $file]} {
         if {![file exists $file]} {
            continue
         }
         set val [gainControl::create $file]
         if {$i%10==0} {
            db::commit
         }
         incr done
      }
      incr i
   }
   #soundControl::latency
   waitWindow::close
}
##
# Return the current play position
#
proc soundControl::position {args} {
   variable startOffset
   return [expr {$startOffset + [snack::audio elapsedTime]}]
}
proc soundControl::remaining {args} {
   return [expr {round($gui::songDuration - $gui::trackPosition)}]
}
proc soundControl::samplePosition {args} {
   variable soundObj
   set rate [$soundObj cget -rate]
   return [expr {round([position]*$rate)}]
}
proc soundControl::sound {{fname ""}} {
   variable soundObj
   if {[string length $fname] > 0 && [file exists $fname]} {
      $soundObj config -file $fname -fileformat [getType $fname]
   }
   return $soundObj
}
proc soundControl::infoObj {{fname ""}} {
   variable infoObj
   if {[string length $fname] > 0 && [file exists $fname]} {
      $infoObj configure -file $fname -fileformat [getType $fname]
   }
   return $infoObj
}
proc soundControl::fileName {{name "xxx"}} {
   variable soundObj
   variable Trace
   if {$name ne "xxx"} {
      $soundObj configure -file $name -debug $Trace -fileformat ""
   }
   set fn [$soundObj cget -file]
   if {$fn eq ""} {
      # Streams return blank, this will get their shortcut
      set fn $::TA(currentMF)
   }
   return $fn
}
#------------------------------------------------------------------------------
# Function   : isPlaying
# Description: Determine if a song is playing or not
# Author     : Tom Wilkason
# Date       : 10/30/2001
#------------------------------------------------------------------------------
proc soundControl::isPlaying {args} {
   variable playState
   # Note audo active is true till the latency period expires
   # which means the buffer is flushed. We want to know when
   # the buffer has been filled, not flushed.
   if {[string equal $playState "Playing"]} {
      return 1
   } else {
      return 0
   }
}
#------------------------------------------------------------------------------
# Function   : soundControl::pauseRelief
# Description: If paused, push the pause button up/down. Called from pause/stop
#              If stopped, then disable the button
# Author     : Tom Wilkason
#------------------------------------------------------------------------------
proc soundControl::pauseRelief {down} {
   Trace
   global TA snackAmpSettings
   variable butState
   variable butRelief
   #$TA(pauseButton) configure -state $butState($down) ;#-relief $butRelief($down)
   if {$down==1} {
      $TA(pauseButton) configure  \
         -relief sunken -background [imageTools::brighten $snackAmpSettings(bg) 15]
         [$TA(pauseButton) cget -image] configure -gamma 0.5
   } else {
      $TA(pauseButton) configure  \
         -relief flat -background $snackAmpSettings(bg)
         [$TA(pauseButton) cget -image] configure -gamma 1
   }
   if {[winfo exists .miniPlayer]} {
      .miniPlayer.frame.mini_pause configure \
         -state $butState($down) -relief $butRelief($down)
   }
   if {[winfo exists .albumart]} {
      .albumart.pc.mini_pause configure \
         -state $butState($down) -relief $butRelief($down)
   }
}
#------------------------------------------------------------------------------
# Function   : soundControl::pause
# Description: Toggle pausing of a sound object
# Author     : Tom Wilkason
#------------------------------------------------------------------------------
proc soundControl::pause {how} {
   Trace
   variable paused
   variable playState
   if {[snack::audio active]} {
      pauseRelief $how
      snack::audio pause
      set paused [expr {!$paused}]
      if {$paused} {
         set playState "Paused"
      } else {
         set playState "Playing"
      }
   }
}
##
# Function to stop sound and return after it is verifyied stopped
#
proc soundControl::stop {args} {
   Trace
   variable playState
   variable soundObj
   variable paused
   if {[snack::audio active]} {
      pauseRelief 2
      $soundObj stop
   }
   set paused 0
   set playState "Stopped"
}
##
# Manually reset the sound object to clear any open files
# Bind this to stop keys instead of Stop
#
proc soundControl::resetObj {args} {
   Trace
   variable soundObj
   if {[info command $soundObj] ne ""} {
      $soundObj stop
      while {[snack::audio elapsedTime] > 0.0} {
         after 20
         Update
      }
      $soundObj destroy
   }
   set soundObj [snack::sound]
}
#----------------------------------------------------------------------------
# Restore the old title, called after updating it
#----------------------------------------------------------------------------
proc soundControl::restoreTitle {args} {
   variable Title
   set ::songArray(Title) $Title
}

proc soundControl::pushTitle {tit} {
   variable Title $tit
}
#------------------------------------------------------------------------------
# Function   : soundControl::play
# Description: Encapsulate the play
#              Called from the command line.
# Author     : Tom Wilkason
# Date       : 11/18/2001
#------------------------------------------------------------------------------
proc soundControl::play {xx start args} {
   Trace
   variable playState
   variable paused
   variable startOffset
   variable soundObj
   variable lastRate
   variable Title
   set Title $::songArray(Title)

   ##
   # In snack 2.2, the output buffer is kept full,
   # however at about 6 hours expr pow(2,31)/(2*44100*3600)=6.7
   # an internal int counter rolls over and causes the stream
   # to stop. So we simply block until the
   # output buffer flushes to cleanly restart
   # the counter.
   #
   # There is also a bug with globalRate, if a track change occurs
   # and there is a different sampling rate, a different path in
   # calling getsample is used, which causes a bunch of re-seeks.
   # the workaround is to let the buffer fully flush first.
   set myRate [$soundObj cget -rate]
   if {$myRate != $lastRate} {
      set maxSec 0.0
   } else {
      set maxSec 21600.0
   }
#   set now [clock clicks -milliseconds]
   while {[snack::audio elapsedTime] > $maxSec} {
      after 20
      Update
   }
#   set then [clock clicks -milliseconds]
#   saLog "Stalled for [expr {$then-$now}] milliseconds"
   set lastRate $myRate ;# save rate
   set end [gainControl::getStop [soundControl::fileName] $soundObj]
#puts "front=[expr {$start/44100.}] end=[expr {([$soundObj length]-$end)/44100.}]"
   if {[catch {eval $soundObj play -start $start -end $end $args} result] } then {
      saLog "$result\n$::errorInfo"
      tk_messageBox -icon warning -type ok -message "$result\n$::errorInfo"
   } else {
      pauseRelief 0
      set paused 0
      set playState "Playing"
   }
   # Save position in seconds of start of song, offset it by any elapsedTime
   # that may be active from a queued play. Typically elapsedTime will be 0.
   set startOffset [expr {double($start)/double([$soundObj cget -rate])-[snack::audio elapsedTime]}]
   #puts "$startOffset = $start / [$soundObj cget -rate] - [snack::audio elapsedTime]"
}
#------------------------------------------------------------------------------
# Initialize the mixer controls
# will recouple the volume to an external mixer
#------------------------------------------------------------------------------
proc soundControl::initVolume {args} {
   Trace
   global snackAmpSettings
   if {$snackAmpSettings(globalMixer)} {
      # Only init if we are controlling mixer
      foreach {line} [snack::mixer lines] {
         set ::TA(line,$line) [snack::mixer volume $line]
         if {[snack::mixer channels $line] == "Mono"} {
            snack::mixer volume $line ::TA(line,$line)
         } else {
            snack::mixer volume $line ::TA(line,$line) ::TA(line,$line)
         }
      }
      set soundControl::acGaindB $snackAmpSettings(lastGlobalGaindB)
   } else {
      soundControl::setTrackGain 1.0
      soundControl::setGlobalGain $snackAmpSettings(lastGlobalGaindB)
   }
}
#------------------------------------------------------------------------------
# set the global gain factors, typically called from vol slider
# db is from -80 to 0
#------------------------------------------------------------------------------
proc soundControl::setGlobalGain {db} {
   #Trace 1
   variable globalGain
   variable trackGain
   variable lineSelect
   variable lastGlobalGaindB
   variable acGaindB
   variable globalMixer
   set db [clamp $db $::snackAmpSettings(minDB) 0]
   if {$globalMixer} {
      if {$::snackAmpSettings(linscale)} {
         set ::TA(line,$lineSelect) [expr {[fromlinDb $db]*100.0}]
      } else {
         set ::TA(line,$lineSelect) [expr {[fromlogDb $db]*100.0}]
      }
      set lastGlobalGaindB $db
      set acGaindB $db
   } else {
      set globalGain [fromlogDb $db]
      set newGain [clamp [expr {$globalGain * $trackGain}] 0 1.0]
      snack::audio scaling $newGain        ;# need 0-1.0
      set acGaindB $db
      set lastGlobalGaindB $db
   }
}
#----------------------------------------------------------------------------
# Return gain for current track
# return is from 0 to 100
#----------------------------------------------------------------------------
proc soundControl::getGlobalGain {args} {
   variable globalGain
   variable globalMixer
   variable lineSelect
   variable acGaindB
   if {$globalMixer} {
      #set gain $::TA(line,$lineSelect)
      set gain [expr {[fromlogDb $acGaindB]*100.0}]
   } else {
      set gain [expr {$globalGain*100.0}]
   }
   return $gain
}

#----------------------------------------------------------------------------
# Set the gain for a specific track
# percent is 0 to 1.0
#----------------------------------------------------------------------------
proc soundControl::setTrackGain {percent} {
   Trace
   variable globalGain
   variable globalMixer
   variable trackGain
   set percent [clamp $percent 0 5.0]
   if {$globalMixer} {
      snack::audio scaling $percent
   } else {
      set trackGain $percent
      set newGain [clamp [expr {$globalGain * $trackGain}] 0.0 1.0]
      snack::audio scaling $newGain
   }
}
#----------------------------------------------------------------------------
# Return gain for current track
# return is from 0 to 1.0
#----------------------------------------------------------------------------
proc soundControl::getTrackGain {args} {
   variable trackGain
   variable globalMixer
   if {$globalMixer} {
      return [snack::audio scaling]
   } else {
      return $trackGain
   }
}
#------------------------------------------------------------------------------
# Return the snack compatable file type to speed up file type detection
#------------------------------------------------------------------------------
proc soundControl::getType {file} {
   return [string trim [file extension $file] .]
}
#------------------------------------------------------------------------------
# Stop playing the track and return a flag to resume play
# to be used when modifying a file
#------------------------------------------------------------------------------
proc soundControl::isPlayingTrack {file _location} {
   upvar $_location location
   ## TODO: Make this a routine to reuse elsewhere
   set currentSound [soundControl::fileName]
   if {($currentSound eq $file && ![file isdirectory $file]) || ([string match "$file*" $currentSound] && [file isdirectory $file])} {
      # This is a trick to see if the song is actually playing
      set wasPlaying [soundControl::isPlaying]
      if {$wasPlaying} {
         set location [soundControl::samplePosition]
      } else {
         set location 0
      }
      soundControl::stop
      soundControl::resetObj
      set resumePlay 1
   } else {
      set resumePlay 0
   }
   return $resumePlay
}
#----------------------------------------------------------------------------
# Resume playing a track, typically after it has been renamed or modified
#----------------------------------------------------------------------------
proc soundControl::resumePlayingTrack {file location} {
   soundControl::fileName $file
   if {$location > 0} {
      # Resume play after calling routines finish
      after 100 Play current $location
   }
}
#------------------------------------------------------------------------------
# Function   : soundControl::logPower
# Description: Compute the averge powerlevel of the song for level equalization
# Author     : Tom Wilkason
# Date       : 11/10/2001
#------------------------------------------------------------------------------
proc soundControl::logPower {snd {seconds 10} {start 10000} {stop 10000}} {
   Trace
   variable autoLevelReference
   variable autoLevelWindow
   variable gainFactor
   variable autoLevelMaxSamples

   # Power command works like this:
   # windowlengh is area within each block to keep a level^2 total
   # for each sample in each window, power is squared and accumlated
   # Number of blocks to do is the end-start samples divded by the frame
   # length. The frame lengh is specificed in seconds, but internally
   # converted to samples. For example
   # start = 100,000, end = 320,000 (in samples), framelength = 1 (44.1k samples)
   # A total of 220,000/44100 5 blocks will be spaced in region. Each block
   # will be computed on starting at the beginning of each block for the
   # window length (e.g. initial 8192 samples of a 44000 sample block)
   # Start after any possible header (framelen is in seconds, window is
   # put at start of each frame. num frames is total width/frame len (in samples)
   # at 10 seconds, or 441,000 samples. A window length of 8192 will sample ~ 2%
   set trackLen [$snd length -units seconds]
   set samples [expr {$trackLen/double($seconds)}]
   # Make 60 a setting, also the 68 below (along with snack debug level)
   set max $autoLevelMaxSamples
   if {$samples > $max} {
      # Scale back time if over 60 samples
      set seconds [expr {$seconds*double($samples)/$max}]
   }
   set Totals [list]
   set levels [list]
   if {[catch {
   # If we use a rectangle window, the variance is bigger so to use it
   # we need to adjust the scaling again.
      set levels [$snd power -start $start -end [expr {[$snd length]-$stop}] \
         -framelength $seconds \
         -windowlength $autoLevelWindow]
      foreach level $levels {
         if {[string is double -strict $level]} {
            lappend Totals $level
         }
      }
      # Sort so we can toss out low samples, then take the average of the rest
      set llTotals [llength $Totals]
      if {$llTotals > 0} {
         set Totals [lsort -decreasing -real $Totals]
#puts $Totals
         # Grab the top ones for consideration (we are concerned with max power)
         if {$llTotals > 12} {
            set end 4 ;#5
         } elseif {$llTotals > 6} {
            set end 2
         } else {
            set end [expr {$llTotals/2}]
         }
         set slope [gainControl::slope [lsort -real $Totals]]
         set Totals [lrange $Totals 0 $end]
         # Don't brack expr below
         set average [expr ([join $Totals +]) / double([llength $Totals])]
         # Compute offset, 6dB is a double. 68dB is a reference for song that won't clip
         # The absolute peak is about 80 but will only find it if entire file is scanned.
         set offset [expr {pow(10.,(68.0 + $autoLevelReference - $average)/$gainFactor)}]   ;# 20 is theory, 33 doesn't expand as much
#puts "Dynamics=$slope Gain=$offset->[file tail [$snd cget -file]]"
      } else {
         set offset 1.0
      }
      # Round to 2 decimal places
      set value [expr {round($offset*1000.)/1000.}]
   } result]} {
      saLog "soundControl::logPower Error: [$snd cget -file]\n$levels\n$result\n$::errorInfo"
      set value 1.0
   }
   return $value
}
##
# Compute the start point of the audible sound on track
# used to skip over blank sections in a track
#
proc soundControl::findStart {snd} {
   Trace
   variable autoLevelInc
   variable Thresh
   set sPerFFT 20  ;# ~ 0.5 sec
   set Totals [list]
   set sndLength [$snd length]
   set srate [$snd cget -rate]
   set blockLen [expr {round($autoLevelInc*$srate*$sPerFFT)}]
   set wl [expr {round($blockLen*$autoLevelInc)}]
   set startPos 0
   set endPos $blockLen
   while {$startPos < $sndLength} {
      set levels [$snd power -windowtype rectangle -start $startPos -end $endPos -framelength $autoLevelInc -windowlength $wl]
      set i 0
#puts $levels
      foreach {level} $levels {
         if {$level > $Thresh} {
            set pos [expr {round($startPos + $autoLevelInc*$srate*$i)}]
#puts "Start: from beg=[HMSfrac [expr {$pos/double($srate)}]] $pos"
            return $pos
         }
         incr i
      }
      Update
      incr startPos $blockLen
      incr endPos $blockLen
   }
   return $startPos
}
##
# Find the last position in a file which contains audible sound
# used to skip over blank sections in a track
proc soundControl::findStop {snd} {
   Trace
   variable autoLevelInc
   variable Thresh
   set Totals [list]
   set sPerFFT 40  ;# ~1 second
   set endPos [$snd length]
   set srate [$snd cget -rate]
   set blockLen [expr {round($autoLevelInc*$srate*$sPerFFT)}]
   set wl [expr {round($blockLen*$autoLevelInc)}]
   while {$endPos > 0} {
      set startPos [expr {$endPos-$blockLen}]
      set levels [lflip [$snd power -windowtype rectangle -start $startPos -end $endPos -framelength $autoLevelInc -windowlength $wl]]
#puts [lflip $levels]
      set i [llength $levels]
      foreach {level} $levels {
         if {$level > $Thresh} {
            set pos [expr {round($startPos + $autoLevelInc*$srate*($i+1))}]
#puts "$startPos + $autoLevelInc*$srate*$i"
#puts "Stop: from end=[HMSfrac [expr {([$snd length] - $pos)/double($srate)}]] $pos"
            return $pos
         }
         incr i -1
      }
      incr endPos -$blockLen
      Update
   }
   return $endPos
}
package provide soundControl 1.0
